import urllib

import fingerprints
import numpy
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
import splink.duckdb.comparison_template_library as ctl
from pyarrow import json  # noqa: F401
from splink.duckdb.blocking_rule_library import block_on
from splink.duckdb.comparison_library import exact_match
from splink.duckdb.linker import DuckDBLinker


def dedupe_organizations(path_organization_json: str, path_output: str):
    # Load the .jsonl file produced by the parser to parquet table
    table = pa.json.read_json(path_organization_json)

    # Load df from table
    df = table.to_pandas()

    # Normalise 'identifier' column
    identifier = pd.json_normalize(df["identifier"])
    identifier = identifier.add_prefix("identifier_")

    # Normalise 'address' column
    address = pd.json_normalize(df["address"])
    address = address.add_prefix("address_")

    # Normalise 'contactPoint' column
    contactPoint = pd.json_normalize(df["contactPoint"])
    contactPoint = contactPoint.add_prefix("contactPoint_")

    # Normalise 'details' column
    details = pd.json_normalize(df["details"])
    details = details.add_prefix("details_")

    # Concatenate the normalised columns to the original df
    df = pd.concat([df, identifier, address, contactPoint, details], axis=1)

    # Some more mangling
    df["roles"] = df["roles"].apply(lambda x: x[0] if x.any() else None)
    df["details_classification_tedcatype"] = df["details_classifications"].apply(
        lambda x: x[0]["id"] if isinstance(x, numpy.ndarray) and x.any() else None
    )
    df["details_classification_cofog"] = df["details_classifications"].apply(
        lambda x: x[1]["id"] if isinstance(x, numpy.ndarray) and len(x) > 1 else None
    )

    # Normalize the 'name' column using fingerprints
    df["name"] = df["name"].apply(lambda x: fingerprints.generate(x))
    # Case: Dataport AÖR
    df["name"] = df["name"].apply(
        lambda x: "dataport aor" if isinstance(x, str) and "dataport" in x else x
    )

    # Clean the 'details_url' column using urllib.parse and only return hostname and path
    df["details_url"] = df["details_url"].apply(
        lambda x: None if not isinstance(x, str) else urllib.parse.urlparse(x).hostname
    )

    # Drop the original columns
    df = df.drop(
        [
            "identifier",
            "address",
            "contactPoint",
            "details",
            "additionalIdentifiers",
            "contactPoint_url",
            "identifier_scheme",
            "identifier_legalName",
            "details_classifications",
        ],
        axis=1,
    )

    # Replace all None values with NaN
    df = df.replace({None: numpy.nan})

    # Set up the linker
    settings = {
        "link_type": "dedupe_only",
        "unique_id_column_name": "id",
        "blocking_rules_to_generate_predictions": [
            block_on("name"),
            block_on("substr(name, 1, 12)"),
            block_on("substr(contactPoint_email, 1, 10)"),
            block_on("contactPoint_name"),
            block_on("contactPoint_telephone"),
            block_on("contactPoint_faxNumber"),
            block_on("details_url"),
            block_on(["substr(address_streetAddress, 1, 6)", "address_region"]),
            block_on(["identifier_id", "address_countryName"]),
            block_on(["substr(name, 1, 6)", "address_region"]),
            block_on(
                [
                    "details_classification_tedcatype",
                    "details_classification_cofog",
                    "address_region",
                ]
            ),
        ],
        "comparisons": [
            ctl.name_comparison("name", term_frequency_adjustments=True),
            ctl.name_comparison("contactPoint_name", term_frequency_adjustments=True),
            ctl.email_comparison("contactPoint_email", include_domain_match_level=True),
            ctl.postcode_comparison("address_postalCode"),
            exact_match("address_streetAddress"),
            exact_match("identifier_id"),
            exact_match("details_url"),
            exact_match(
                "details_classification_tedcatype", term_frequency_adjustments=True
            ),
            exact_match(
                "details_classification_cofog", term_frequency_adjustments=True
            ),
        ],
    }
    linker = DuckDBLinker(df, settings)

    # Estimate the parameters of the model
    linker.estimate_probability_two_random_records_match(
        [block_on(["contactPoint_name"])], recall=0.8
    )
    linker.estimate_u_using_random_sampling(max_pairs=1e7)
    linker.estimate_parameters_using_expectation_maximisation(
        block_on(["contactPoint_name"]),
        estimate_without_term_frequencies=True,
    )

    # Predict the matches and convert the mathces with a probability above 0.2
    # to a pandas dataframe
    df_predict = linker.predict(threshold_match_probability=0.2)
    df_predict = df_predict.as_pandas_dataframe()

    # Create a new dataframe with only the needed columns
    df_result = df_predict[["id_l", "id_r", "match_probability"]]

    # Rename those columns
    df_result = df_result.rename(columns={"id_l": "from", "id_r": "to"})

    # Remove all rows where match_probability is below 0.5 and remove the index
    df_result = df_result[df_result["match_probability"] > 0.1]
    df_result = df_result.reset_index(drop=True)

    # Write df_result to a parquet file in the output folder and name it
    # 'edges_isSameAs.parquet'
    table = pa.Table.from_pandas(df_result)
    pq.write_table(table, f"{path_output}/edges_isSameAs.parquet")



# SPDX-FileCopyrightText: 2023 Free Software Foundation Europe <contact@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Dict, Optional

import ftfy
import phonenumbers
from _decimal import Decimal
from lxml import etree

from tedective_etl.schema import ocds
from tedective_etl.utils import clean, convert_to_euro


def extract_title(
    elem: etree._Element,
    doc_sec_trans: etree._Element,
) -> Optional[str]:
    if elem.find(".//TITLE", namespaces=elem.nsmap) is None:
        return [
            doc.find(".//TI_TEXT", namespaces=elem.nsmap)[0].text
            for doc in doc_sec_trans.iterfind(".//ML_TI_DOC", namespaces=elem.nsmap)
            if doc.get("LG") == "EN"
        ][0]

    orig_title = elem.find(".//TITLE", namespaces=elem.nsmap).text
    if not orig_title:
        orig_title = elem.find(".//TITLE", namespaces=elem.nsmap)[0].text

    if doc_sec_trans is None:
        return clean(orig_title)

    else:
        try:
            title = (
                [
                    doc.find(".//TI_TEXT", namespaces=elem.nsmap)[0].text
                    for doc in doc_sec_trans.iterfind(
                        ".//ML_TI_DOC", namespaces=elem.nsmap
                    )
                    if doc.get("LG") == "EN"
                ][0]
                + " ["
                + orig_title
                + "]"
            )
            return clean(title)
        except TypeError:
            return orig_title


def extract_description(elem: etree._Element) -> str:
    description = elem.find(".//SHORT_DESCR", namespaces=elem.nsmap)[0].text
    return clean(description)


def extract_text(elem: etree._Element, path: str) -> Optional[str]:
    try:
        to_return = elem.find(path, namespaces=elem.nsmap).text
        return ftfy.fix_text(to_return)
    except TypeError:
        to_return = elem.find(path, namespaces=elem.nsmap)[0].text
        try:
            return ftfy.fix_text(to_return)
        except TypeError:
            return None
    except AttributeError:
        return None


def extract_buyer_contact_point(
    elem: etree._Element, buyer_search_str: Optional[str] = "ADDRESS_CONTRACTING_BODY"
) -> ocds.ContactPoint:
    name = extract_text(elem, f".//CONTRACTING_BODY/{buyer_search_str}/CONTACT_POINT")
    phone_unformatted = extract_text(
        elem, f".//CONTRACTING_BODY/{buyer_search_str}/PHONE"
    )
    if phone_unformatted:
        if "/" in phone_unformatted:
            phone_unformatted = phone_unformatted.split("/")[0]
        try:
            phonenumber = phonenumbers.parse(phone_unformatted)
            phone = phonenumbers.format_number(
                phonenumber, phonenumbers.PhoneNumberFormat.INTERNATIONAL
            )
        except phonenumbers.phonenumberutil.NumberParseException:
            phone = phone_unformatted + " [likely invalid]"
    else:
        phone = None
    email = extract_text(elem, f".//CONTRACTING_BODY/{buyer_search_str}/E_MAIL")
    fax_unformatted = extract_text(elem, f".//CONTRACTING_BODY/{buyer_search_str}/FAX")
    if fax_unformatted:
        try:
            faxnumber = phonenumbers.parse(fax_unformatted)
            fax = phonenumbers.format_number(
                faxnumber, phonenumbers.PhoneNumberFormat.INTERNATIONAL
            )
        except phonenumbers.phonenumberutil.NumberParseException:
            fax = fax_unformatted + " [likely invalid]"
    else:
        fax = None

    url = extract_text(elem, f".//CONTRACTING_BODY/{buyer_search_str}/URL_GENERAL")
    return ocds.ContactPoint(
        name=name, telephone=phone, email=email, faxNumber=fax, url=url
    )


def extract_buyer_details(
    elem: etree._Element, buyer_search_str: Optional[str] = "ADDRESS_CONTRACTING_BODY"
) -> Dict:
    url = extract_text(elem, f".//CONTRACTING_BODY/{buyer_search_str}/URL_GENERAL")
    buyerProfile = extract_text(
        elem, f".//CONTRACTING_BODY/{buyer_search_str}/URL_BUYER"
    )
    return {"url": url, "buyerProfile": buyerProfile}


def extract_buyer_address(
    elem: etree._Element,
    base_path: Optional[str] = "CONTRACTING_BODY",
    search_path: Optional[str] = "ADDRESS_CONTRACTING_BODY",
) -> ocds.Address:
    # TODO Think about using libpostal here for normalizing
    streetAddress = extract_text(elem, f".//{base_path}/{search_path}/ADDRESS")
    locality = extract_text(elem, f".//{base_path}/{search_path}/TOWN")
    postalCode = extract_text(elem, f".//{base_path}/{search_path}/POSTAL_CODE")

    try:
        region = elem.find(
            f".//{base_path}/{search_path}" + "/{*}NUTS", namespaces=elem.nsmap
        ).get("CODE")
    except AttributeError:
        region = None

    try:
        countryName = elem.find(
            f".//{base_path}/{search_path}/COUNTRY",
            namespaces=elem.nsmap,
        ).get("VALUE")
    except AttributeError:
        # TODO Do proper mapping to country names here
        if region:
            countryName = region[:2]
        else:
            countryName = None

    return ocds.Address(
        streetAddress=streetAddress,
        locality=locality,
        region=region,
        postalCode=postalCode,
        countryName=countryName,
    )


def extract_supplier_address(
    elem: etree._Element,
) -> ocds.Address:
    # TODO Think about using libpostal here for normalizing
    streetAddress = extract_text(elem, ".//ADDRESS")
    locality = extract_text(elem, ".//TOWN")
    postalCode = extract_text(elem, ".//POSTAL_CODE")

    try:
        region = elem.find(".//{*}NUTS", namespaces=elem.nsmap).get("CODE")
    except AttributeError:
        region = None

    try:
        countryName = elem.find(
            ".//COUNTRY",
            namespaces=elem.nsmap,
        ).get("VALUE")
    except AttributeError:
        if region:
            countryName = region[:2]
        else:
            countryName = None

    return ocds.Address(
        streetAddress=streetAddress,
        locality=locality,
        region=region,
        postalCode=postalCode,
        countryName=countryName,
    )


def _extract_value(elem: etree._Element, path: str, date) -> Optional[ocds.Value]:
    amount = extract_text(elem, path)
    currency = extract_attribute(elem, path, "CURRENCY")
    amountEur = None
    if amount is not None:
        if currency != "EUR":
            amountEur = convert_to_euro(amount=amount, currency=currency, date=date)
        else:
            # amount = float(Decimal(amount))
            amountEur = amount
        return ocds.Value(
            amount=float(Decimal(amount)),
            currency=ocds.Currency[currency],
            amountEur=amountEur,
        )
    else:
        return None


def extract_value(elem: etree._Element, ocds_object: str, date: str = None) -> Optional[ocds.Value]:
    # TODO Update currency conversion without re-ingesting all the data

    if ocds_object == "tender":
        tender_value_paths = [
            ".//OBJECT_CONTRACT/VAL_ESTIMATED_TOTAL",
        ]

        for path in tender_value_paths:
            value = _extract_value(elem, path, date)
            if value is not None:
                return value
            else:
                continue

    elif ocds_object == "award":
        award_value_paths = [
            ".//VAL_TOTAL",
            ".//VALUES/VAL_TOTAL",
        ]

        for path in award_value_paths:
            value = _extract_value(elem, path, date)
            if value is not None:
                return value
            else:
                continue


def extract_attribute(elem: etree._Element, path: str, attribute: str) -> Optional[str]:
    try:
        to_return = elem.find(path, namespaces=elem.nsmap).get(attribute)
        return ftfy.fix_text(to_return)
    except AttributeError:
        return None



# SPDX-FileCopyrightText: 2022 Free Software Foundation Europe <contact@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import uuid
from datetime import datetime
from typing import List, Optional

import fingerprints
from lxml import etree

from tedective_etl.schema import ocds
from tedective_etl.ted_to_ocds import extractors, trackers

implemented_forms = ["F01", "F02", "F03"]


def ted_notice_to_ocds_releases(
    doc_id: str,
    ted_url: str,
    form_type: str,
    ocid_prefix: str,
    object_contract: List[etree._Element],
    doc_sec_coded: etree._Element,
    doc_sec_forms: etree._Element,
    doc_sec_trans: etree._Element,
    translate_from: Optional[str] = None,
) -> tuple():
    """
    Returns a list of OCDS releases from the XML source of TED notices.
    See https://standard.open-contracting.org/profiles/eu/latest/en/forms/F01/
    for more information.

    :param form_type:
    :param ocid_prefix:
    :param doc_id:
    :param ted_url:
    :param object_contract:
    :param doc_sec_forms:
    :param doc_sec_coded:
    :param doc_sec_trans:
    :param translate_from:
    :rtype:
    """
    releases = []
    organizations = []

    # Only process implemented form types
    if form_type not in implemented_forms:
        return releases, organizations

    # A simple counter to keep track of the number of elements in one '/OBJECT_CONTRACT'
    elem_count = 0

    # Generate an OCDS release for each '/OBJECT_CONTRACT' element
    elem: etree._Element
    for elem in object_contract:
        elem_count += 1

        # Generate OCID
        if form_type == "F01" and len(object_contract) == 1:
            # If it's an F01 form and there is only one '/OBJECT_CONTRACT' element,
            # generate an OCID based on 'DOC_ID'
            id = doc_id
            ocid = ocid_prefix + str(uuid.uuid5(uuid.NAMESPACE_URL, doc_id))
        elif form_type == "F01" and len(object_contract) > 1:
            # There are more than one '/OBJECT_CONRACT' elements, for each of which a
            # new release with a new OCID needs to be generated, so use the counter
            id = f"{doc_id}-{elem_count}"
            ocid = ocid_prefix + str(
                uuid.uuid5(uuid.NAMESPACE_URL, f"{doc_id}-{elem_count}")
            )
        elif form_type == "F02" or form_type == "F03":
            # If it's an F02 form or an F03 form, first check whether there are
            # related documents.
            related_doc_id = extractors.extract_text(
                doc_sec_forms, ".//PROCEDURE/NOTICE_NUMBER_OJ"
            )
            if related_doc_id is not None:
                trackers.related_doc_id_count += 1
            # If there are related documents, generate an OCID based on the
            # existing document ID
            if related_doc_id is not None:
                if len(object_contract) > 1:
                    id = f"{doc_id}-{elem_count}"
                    ocid = ocid_prefix + str(
                        uuid.uuid5(uuid.NAMESPACE_URL, f"{related_doc_id}-{elem_count}")
                    )
                else:
                    id = doc_id
                    ocid = ocid_prefix + str(
                        uuid.uuid5(uuid.NAMESPACE_URL, related_doc_id)
                    )
            else:
                if len(object_contract) > 1:
                    id = f"{doc_id}-{elem_count}"
                    ocid = ocid_prefix + str(
                        uuid.uuid5(uuid.NAMESPACE_URL, f"{doc_id}-{elem_count}")
                    )
                else:
                    id = doc_id
                    ocid = ocid_prefix + str(uuid.uuid5(uuid.NAMESPACE_URL, doc_id))
        else:
            id = doc_id
            ocid = ocid_prefix

        # Retrieve date when Release was published
        date =  datetime.strptime(
                    doc_sec_coded.find(".//DATE_PUB", namespaces=doc_sec_coded.nsmap).text,
                    "%Y%m%d",  # type: ignore
        )

        # Define Release object instance and set some initial values
        release = ocds.Release(
            id=id,
            ocid=ocid,
            language="en",
            date=date,
            initiationType=ocds.InitiationType.tender,
            tag=[ocds.TagEnum.tender],
            tedURL=ted_url,
        )

        # Define linked objects and optionally set some defaults
        # II.1.1
        # tender_id = extractors.extract_text(elem, ".//REFERENCE_NUMBER")
        # if not tender_id:
        #     tender_id = release.id
        tender = ocds.Tender(
            # TODO This should be checked for tender ID: /OBJECT_CONTRACT/REFERENCE_NUMBER
            id=str(uuid.uuid4()),
            legalBasis=ocds.Classification(scheme="CELEX"),
        )
        release.tender = tender

        # F01
        if form_type == "F01":
            # https://standard.open-contracting.org/profiles/eu/latest/en/forms/F01/#preamble
            notice_type = doc_sec_forms.find(  # type: ignore
                ".//NOTICE", namespaces=doc_sec_forms.nsmap
            ).get("TYPE")
            if notice_type == "PRI_ONLY" or notice_type == "PRI_REDUCING_TIME_LIMITS":
                release.tag = [ocds.TagEnum.planning]
                tender.status = ocds.Status.planned
            if notice_type == "PRI_CALL_COMPETITION":
                release.tag = [ocds.TagEnum.planning, ocds.TagEnum.tender]
                tender.status = ocds.Status.active

        # F02
        elif form_type == "F02":
            tender.status = ocds.Status.active

        # F03
        elif form_type == "F03":
            release.tag = [ocds.TagEnum.award, ocds.TagEnum.contract]
            tender.status = ocds.Status.complete

        # Section I - Contracting Authority (CA)
        # https://standard.open-contracting.org/profiles/eu/latest/en/forms/F01/#section-i
        # I.1 Name and addresses
        buyer_name = extractors.extract_text(
            doc_sec_forms,
            ".//CONTRACTING_BODY/ADDRESS_CONTRACTING_BODY/OFFICIALNAME",
        )

        buyer_identifier = ocds.Identifier(legalName=buyer_name)
        buyer_identifier.id = extractors.extract_text(
            doc_sec_forms,
            ".//CONTRACTING_BODY/ADDRESS_CONTRACTING_BODY/NATIONALID",
        )
        buyer_identifier.scheme = "National-ID"

        buyer_address = extractors.extract_buyer_address(doc_sec_forms)
        buyer_contactPoint = extractors.extract_buyer_contact_point(doc_sec_forms)
        buyer_details = extractors.extract_buyer_details(doc_sec_forms)

        # Generate a heavily-scrubbed fingerprint from the organization's name
        buyer_fingerprint = fingerprints.generate(buyer_name)
        # Generate UUID based on that fingerprint. Same fingerprint, same UUID.
        try:
            buyer_uuid = uuid.uuid5(uuid.NAMESPACE_URL, buyer_fingerprint)
        except TypeError:
            buyer_uuid = uuid.uuid4()
        buyer = ocds.Organization(
            id=str(buyer_uuid),
            name=buyer_name,
            identifier=buyer_identifier,
            address=buyer_address,
            contactPoint=buyer_contactPoint,
            details=buyer_details,
            roles={"buyer"},
        )
        # Add the organization to the list of organizations
        organizations.append(buyer)

        # Add the buyer to the release
        release.buyer = buyer

        # Add the buyer to the list of parties
        release.parties = set()
        release.parties.add(buyer)

        # Look for additional buyer
        additional_buyer = doc_sec_forms.find(
            ".//CONTRACTING_BODY/ADDRESS_CONTRACTING_BODY_ADDITIONAL",
            namespaces=doc_sec_forms.nsmap,
        )
        if additional_buyer is not None:
            add_buyer_name = extractors.extract_text(
                doc_sec_forms,
                ".//CONTRACTING_BODY/ADDRESS_CONTRACTING_BODY_ADDITIONAL/OFFICIALNAME",
            )

            add_buyer_identifier = ocds.Identifier(legalName=add_buyer_name)
            add_buyer_identifier.id = extractors.extract_text(
                doc_sec_forms,
                ".//CONTRACTING_BODY/ADDRESS_CONTRACTING_BODY_ADDITIONAL/NATIONALID",
            )
            add_buyer_identifier.scheme = "National-ID"

            add_buyer_address = extractors.extract_buyer_address(
                doc_sec_forms, "ADDRESS_CONTRACTING_BODY_ADDITIONAL"
            )
            add_buyer_contactPoint = extractors.extract_buyer_contact_point(
                doc_sec_forms, "ADDRESS_CONTRACTING_BODY_ADDITIONAL"
            )
            # Generate a heavily-scrubbed fingerprint from the organization's name
            add_buyer_fingerprint = fingerprints.generate(add_buyer_name)
            # Generate UUID based on that fingerprint. Same fingerprint, same UUID.
            add_buyer_uuid = uuid.uuid5(uuid.NAMESPACE_URL, add_buyer_fingerprint)
            add_buyer = ocds.Organization(
                id=str(add_buyer_uuid),
                fingerprint=add_buyer_fingerprint,
                name=add_buyer_name,
                identifier=add_buyer_identifier,
                address=add_buyer_address,
                contactPoint=add_buyer_contactPoint,
                roles={"buyer"},
            )
            if add_buyer_uuid == buyer_uuid:
                # Additional buyer is the same as primary buyer, do nothing
                pass
            else:
                # Additional buyer is new, add to lists
                organizations.append(add_buyer)
                release.parties.add(add_buyer)

        # I.2 Information about joint procurement
        cb_pl = doc_sec_forms.find(
            ".//CONTRACTING_BODY/PROCUREMENT_LAW", namespaces=doc_sec_forms.nsmap
        )
        if cb_pl is not None:
            tender.crossBorderLaw = cb_pl[0].text
        cp = doc_sec_forms.find(
            ".//CONTRACTING_BODY/CENTRAL_PURCHASING", namespaces=doc_sec_forms.nsmap
        )
        if cp is not None:
            buyer.roles.append("centralPurchasingBody")

        # TODO I.3 Communication
        cb_url_p = doc_sec_forms.find(
            ".//CONTRACTING_BODY/URL_PARTICIPATION", namespaces=doc_sec_forms.nsmap
        )
        if cb_url_p is not None:
            tender.submissionMethod = ["electronicSubmission"]
            tender.submissionMethodDetails = cb_url_p.text

        classifications = []
        # I.4 Type of the contracting authority
        buyer_details_classification = ocds.Classification(
            scheme="TED_CA_TYPE",
            id=str(
                extractors.extract_attribute(
                    doc_sec_forms, ".//CONTRACTING_BODY/CA_TYPE", "VALUE"
                )
            ),
            # TODO Add description
        )
        if buyer_details_classification.id is not None:
            classifications.append(buyer_details_classification)

        # I.5 Main activity
        buyer_details_main_activity = ocds.Classification(
            scheme="COFOG",
            id=extractors.extract_attribute(
                doc_sec_forms, ".//CONTRACTING_BODY/CA_ACTIVITY", "VALUE"
            ),
        )
        if buyer_details_main_activity.id is not None:
            classifications.append(buyer_details_main_activity)

        if len(classifications) > 0:
            buyer.details["classifications"] = classifications

        # Section II
        # https://standard.open-contracting.org/profiles/eu/latest/en/forms/F01/#section-ii
        # II.1 Scope of the procurement
        # II.1.1 Title
        tender.title = extractors.extract_title(elem, doc_sec_trans)

        # II.1.2 Main CPV Code
        classification = ocds.Classification(
            scheme="CPV", id=extractors.extract_attribute(elem, ".//CPV_CODE", "CODE")
        )
        tender.classification = classification

        # TODO II.1.3
        # II.1.4 Short description
        tender.description = extractors.extract_description(elem)

        # II.1.5 Estimated Total Value
        tender.value = extractors.extract_value(doc_sec_forms, "tender", date)

        # Section V - Award of contract
        if form_type == "F03":
            # Check if there are more awards
            award_elems = doc_sec_forms.findall(
                ".//AWARD_CONTRACT", namespaces=doc_sec_forms.nsmap
            )
            release.awards = set()
            award_elem: etree._Element
            for award_elem in award_elems:
                award_id = uuid.uuid4()
                award_title = extractors.extract_title(award_elem, doc_sec_trans)
                award = ocds.Award(id=str(award_id), title=award_title)

                if extractors.extract_text(award_elem, ".//AWARDED_CONTRACT"):
                    award.status = ocds.AwardStatus.active
                    award_date = datetime.strptime(
                        extractors.extract_text(
                            award_elem,
                            ".//DATE_CONCLUSION_CONTRACT",
                        ),
                        "%Y-%m-%d",
                    )
                    contract = ocds.Contract(
                        id=str(uuid.uuid4()),
                        awardID=str(award_id),
                        title=award_title,
                        description=None,
                        status=ocds.ContractStatus("active"),
                        period=None,
                        value=extractors.extract_value(award_elem, "award", award_date),
                        items=None,
                        dateSigned=award_date,
                        documents=None,
                    )
                    release.contracts = set()
                    release.contracts.add(contract)
                else:
                    award.status = ocds.AwardStatus.unsuccessful

                # V.2.3 Name and address of the contractor(s)
                supplier_elems = award_elem.findall(
                    ".//CONTRACTOR",
                    namespaces=award_elem.nsmap,
                )

                for supplier_elem in supplier_elems:
                    supplier_name = extractors.extract_text(
                        supplier_elem, ".//OFFICIALNAME"
                    )
                    supplier_fingerprint = fingerprints.generate(supplier_name)
                    try:
                        supplier_id = uuid.uuid5(
                            uuid.NAMESPACE_URL, supplier_fingerprint
                        )
                    except TypeError:
                        supplier_id = uuid.uuid4()
                    supplier = ocds.Organization(
                        id=str(supplier_id),
                        name=supplier_name,
                        fingerprint=supplier_fingerprint,
                        identifier=ocds.Identifier(
                            id=extractors.extract_text(supplier_elem, ".//NATIONALID"),
                            scheme="National-ID",
                            legalName=supplier_name,
                        ),
                        roles={"supplier"},
                        address=extractors.extract_supplier_address(supplier_elem),
                    )

                    award.suppliers = set()
                    award.suppliers.add(supplier)

                    # This is a new supplier
                    release.parties.add(supplier)

                    # Add supplier to list of organizations
                    organizations.append(supplier)

                    # iterated through all suppliers for award

                # Add award to release
                release.awards.add(award)
                # iterated through all awards

        if release not in releases:
            releases.append(release)
        # End of for-loop through '/OBJECT_CONTRACT'

    # After having iterated through all '/OBJECT_CONTRACT' elements, return releases
    return releases, organizations



# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe <contact@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import calendar
import os

import orjson
import requests


def api_request(start_date: str, end_date: str) -> dict:
    host = "api.frankfurter.app"
    url = f'https://{host}/{start_date}..{end_date}'
    response = requests.get(url)
    rates = response.json()
    return rates["rates"]


def get_last_day_of_month(date_str):
    year, month = map(int, date_str.split('-'))
    _, last_day = calendar.monthrange(year, month)
    return last_day


def fetch_exchange_rates(start_date: str, end_date: str):
    start_date = start_date + "-01"
    end_date = end_date + "-" + str(get_last_day_of_month(end_date))
    rates = api_request(start_date, end_date)
    
    root = "/tmp" #get_project_root()
    dest = os.path.join(root, "ted_currencies")
    dest_file = os.path.join(dest, "rates.json")

    if not os.path.exists(dest):
        os.makedirs(dest)

    with open(dest_file, "wb") as file:
        file.write(orjson.dumps(rates, option=orjson.OPT_INDENT_2))



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2023 Free Software Foundation Europe <contact@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import tarfile
from calendar import monthrange
from datetime import datetime
from glob import glob
from typing import Any, List, Optional

from dateutil import rrule
from lxml import etree

from tedective_etl.ted_to_ocds.transform import ted_notice_to_ocds_releases

log = logging.getLogger()


def extract_ted_notices(in_dir, year_month):
    """
    Extracts TED XMLs for a given year and month.

    Args:
        in_dir (str): Directory where TED XML files are stored
        year_month (str): Year and month to extract

    Returns:
        None
    """
    log.info(f"Extracting TED XMLs for {year_month}...")
    f_to_extract = tarfile.open(f"{in_dir}/{year_month}.tar.gz")
    f_to_extract.extractall(f"{in_dir}/{year_month}")
    os.remove(f"{in_dir}/{year_month}.tar.gz")


def get_ted_xml_files(
    in_dir: str,
    first_month: str = "2017-01",
    last_month: Optional[str] = None,
) -> dict:
    """
    Generates a dictionary of lists of TED XML filepaths from a given directory

    Args:
        in_dir (str): Directory where TED XML files are stored
        first_month (str): First month to get XML files for
        last_month (str): Last month to get XML files for

    Returns:
        dict: Dictionary of lists of TED XML filepaths from a given directory.
        The key for each list of XML filepaths is the date in the format
        'YYYYMMDD'.
    """
    xml_dict = dict()
    start_date = datetime(
        int(first_month.split("-")[0]),
        int(first_month.split("-")[1]),
        1,
    )

    if last_month is None:
        # Define end_date as the current date
        end_date = datetime.now().strftime("%Y-%m")
    else:
        end_date = last_month

    end_year = int(end_date.split("-")[0])
    end_month = int(end_date.split("-")[1])

    # Get the last day of the month
    _, last_day = monthrange(end_year, end_month)
    end_date = datetime(end_year, end_month, last_day)

    for dt in rrule.rrule(rrule.DAILY, dtstart=start_date, until=end_date):
        year_month = dt.strftime("%Y-%m")
        year_month_day = dt.strftime("%Y%m%d")
        xml_dict[year_month_day] = []
        for filename in glob(f"{in_dir}/{year_month}/{year_month_day}*/*.xml"):
            xml_dict[year_month_day].append(filename)

    # Remove empty lists
    xml_dict = {k: v for k, v in xml_dict.items() if v}

    return xml_dict


def process_ted_notice(xml_filepath: str) -> tuple:
    """
    Returns a dict containing the form type and the list of OCDS releases
    generated from it.

    Args:
        xml_filepath (str): Filepath of the TED XML file to process

    Returns:
        dict: Dictionary containing the form type ('form_type'), the list of
        OCDS releases ('releases'), and the list of organizations
        ('organizations') generated from the TED XML file.
    """

    # Optimize performance for XMLParser
    parser = etree.XMLParser(ns_clean=True, huge_tree=True, remove_blank_text=True)

    # Define tree and generate cleaned root of XML
    tree: etree._ElementTree[Any] = etree.parse(xml_filepath, parser)  # type: ignore
    root: etree._Element = tree.getroot()

    # Define sections based on list indices
    # sec_techn = root[0]
    # sec_links = root[1]
    sec_coded: etree._Element = root[2]
    sec_trans = root[3]
    sec_forms = root[4]

    # This is the notice ID,
    try:
        sec_coded.find(".//NOTICE_DATA/NO_DOC_OJS", namespaces=sec_coded.nsmap).text

    except AttributeError:
        # TODO Implement eForms
        return {"form_type": "eForm", "releases": [], "organizations": []}

    # URI of current TED notice
    doc_id_short = root.get("DOC_ID")
    ted_url = f"https://ted.europa.eu/udl?uri=TED:NOTICE:{doc_id_short}:TEXT:EN:HTML"

    # Define form type
    form_type = sec_forms[0].get("FORM")
    form_version = sec_forms[0].get("VERSION")
    if form_type is None:
        try:
            form_type = sec_forms[1].get("FORM")
            form_version = sec_forms[1].get("VERSION")
            if form_type is None:
                return {"form_type": "invalid", "releases": [], "organizations": []}
        except IndexError:
            if form_version is None:
                return {"form_type": "invalid", "releases": [], "organizations": []}
            else:
                return {"form_type": form_version, "releases": [], "organizations": []}
    if form_type[0].isdigit():
        if len(form_type) == 1:
            form_type = f"F0{form_type}"
        else:
            form_type = f"F{form_type}"

    # Prune 'sec_forms' to only contain English version or if that is not available,
    # another language version. The language from which translation will be needed is
    # also passed
    translate_from = None
    for elem in sec_forms:
        if elem.tag.endswith("NOTICE_UUID"):
            continue
        lang: str = elem.get("LG")
        if lang == "EN":
            new_sec_forms: etree._Element = elem
            # Found an English form; stop iterating
            break
        # No English form was found. Translation will be needed
        new_sec_forms = elem
        translate_from = lang.lower()

    sec_forms = new_sec_forms

    # Check if 'OBJECT_CONTRACT' is indeed a list. If the list contains more than
    # one item, then multiple releases need to be generated.
    object_contract: List = sec_forms.findall(
        ".//OBJECT_CONTRACT", namespaces=sec_forms.nsmap
    )
    assert isinstance(object_contract, List)

    # Pass off to OCDS parser to generate releases
    result = ted_notice_to_ocds_releases(
        # Use doc_id_short as ID as its URL-safe
        doc_id=doc_id_short,
        ted_url=ted_url,
        form_type=form_type,
        ocid_prefix="ocds-jyvdv7-",
        object_contract=object_contract,
        doc_sec_forms=sec_forms,
        doc_sec_coded=sec_coded,
        doc_sec_trans=sec_trans,
        translate_from=translate_from,
    )

    # Construct the dictionary
    processed_ted_notice = {
        "form_type": form_type,
        "releases": result[0],
        "organizations": result[1],
    }
    return processed_ted_notice



# generated by datamodel-codegen:
#   filename:  release-schema.json
#   timestamp: 2023-12-14T08:36:01+00:00

from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Set, Union

from pydantic import AnyUrl, BaseModel, Field, constr


class TagEnum(Enum):
    planning = "planning"
    planningUpdate = "planningUpdate"
    tender = "tender"
    tenderAmendment = "tenderAmendment"
    tenderUpdate = "tenderUpdate"
    tenderCancellation = "tenderCancellation"
    award = "award"
    awardUpdate = "awardUpdate"
    awardCancellation = "awardCancellation"
    contract = "contract"
    contractUpdate = "contractUpdate"
    contractAmendment = "contractAmendment"
    implementation = "implementation"
    implementationUpdate = "implementationUpdate"
    contractTermination = "contractTermination"
    compiled = "compiled"


class InitiationType(Enum):
    tender = "tender"


class Status(Enum):
    planning = "planning"
    planned = "planned"
    active = "active"
    cancelled = "cancelled"
    unsuccessful = "unsuccessful"
    complete = "complete"
    withdrawn = "withdrawn"
    NoneType_None = None


class ProcurementMethod(Enum):
    open = "open"
    selective = "selective"
    limited = "limited"
    direct = "direct"
    NoneType_None = None


class MainProcurementCategory(Enum):
    goods = "goods"
    works = "works"
    services = "services"
    NoneType_None = None


class AwardStatus(Enum):
    pending = "pending"
    active = "active"
    cancelled = "cancelled"
    unsuccessful = "unsuccessful"
    NoneType_None = None


class ContractStatus(Enum):
    pending = "pending"
    active = "active"
    cancelled = "cancelled"
    terminated = "terminated"
    NoneType_None = None


class MilestoneStatus(Enum):
    scheduled = "scheduled"
    met = "met"
    notMet = "notMet"
    partiallyMet = "partiallyMet"
    NoneType_None = None


class Document(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="A local, unique identifier for this document. This field is used to keep track of multiple revisions of a document through the compilation from release to record mechanism.",
        title="ID",
    )
    documentType: Optional[str] = Field(
        None,
        description="A classification of the document described, using the open [documentType](https://standard.open-contracting.org/1.1/en/schema/codelists/#document-type) codelist.",
        title="Document type",
    )
    title: Optional[str] = Field(None, description="The document title.", title="Title")
    description: Optional[str] = Field(
        None,
        description="A short description of the document. Descriptions are recommended to not exceed 250 words. In the event the document is not accessible online, the description field can be used to describe arrangements for obtaining a copy of the document.",
        title="Description",
    )
    url: Optional[AnyUrl] = Field(
        None,
        description="A direct link to the document or attachment. The server providing access to this document ought to be configured to correctly report the document mime type.",
        title="URL",
    )
    datePublished: Optional[datetime] = Field(
        None,
        description="The date on which the document was first published. This is particularly important for legally important documents such as notices of a tender.",
        title="Date published",
    )
    dateModified: Optional[datetime] = Field(
        None,
        description="Date that the document was last modified",
        title="Date modified",
    )
    format: Optional[str] = Field(
        None,
        description="The format of the document, using the open [IANA Media Types](https://www.iana.org/assignments/media-types/media-types.xhtml) codelist (see the values in the 'Template' column), or using the 'offline/print' code if the described document is published offline. For example, web pages have a format of 'text/html'.",
        title="Format",
    )
    language: Optional[str] = Field(
        None,
        description="The language of the linked document using either two-letter [ISO639-1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes), or extended [BCP47 language tags](https://www.w3.org/International/articles/language-tags/). The use of lowercase two-letter codes from [ISO639-1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes) is recommended unless there is a clear user need for distinguishing the language subtype.",
        title="Language",
    )


class Change(BaseModel):
    property: Optional[str] = Field(
        None,
        description="The property name that has been changed relative to the place the amendment is. For example if the contract value has changed, then the property under changes within the contract.amendment would be value.amount. (Deprecated in 1.1)",
        title="Property",
    )
    former_value: Optional[Union[str, float, int, List[Any], Dict[str, Any]]] = Field(
        None,
        description="The previous value of the changed property, in whatever type the property is. (Deprecated in 1.1)",
        title="Former Value",
    )


class Amendment(BaseModel):
    date: Optional[datetime] = Field(
        None, description="The date of this amendment.", title="Amendment date"
    )
    rationale: Optional[str] = Field(
        None, description="An explanation for the amendment.", title="Rationale"
    )
    id: Optional[str] = Field(
        None,
        description="An identifier for this amendment: often the amendment number",
        title="ID",
    )
    description: Optional[str] = Field(
        None,
        description="A free text, or semi-structured, description of the changes made in this amendment.",
        title="Description",
    )
    amendsReleaseID: Optional[str] = Field(
        None,
        description="Provide the identifier (release.id) of the OCDS release (from this contracting process) that provides the values for this contracting process **before** the amendment was made.",
        title="Amended release (identifier)",
    )
    releaseID: Optional[str] = Field(
        None,
        description="Provide the identifier (release.id) of the OCDS release (from this contracting process) that provides the values for this contracting process **after** the amendment was made.",
        title="Amending release (identifier)",
    )
    changes: Optional[List[Change]] = Field(
        None,
        description="An array of change objects describing the fields changed, and their former values. (Deprecated in 1.1)",
        title="Amended fields",
    )


class Classification(BaseModel):
    scheme: Optional[str] = Field(
        None,
        description="The scheme or codelist from which the classification code is taken. For line item classifications, this uses the open [itemClassificationScheme](https://standard.open-contracting.org/1.1/en/schema/codelists/#item-classification-scheme) codelist.",
        title="Scheme",
    )
    id: Optional[Union[str, int]] = Field(
        None, description="The classification code taken from the scheme.", title="ID"
    )
    # TODO This needs to implemented. It's currently not used to simplify the schema for KuzuDB.
    # description: Optional[str] = Field(
    #     None,
    #     description="A textual description or title for the classification code.",
    #     title="Description",
    # )
    # uri: Optional[AnyUrl] = Field(
    #     None,
    #     description="A URI to uniquely identify the classification code.",
    #     title="URI",
    # )

    def __hash__(self) -> int:
        return super().__hash__()


class Identifier(BaseModel):
    scheme: Optional[str] = Field(
        None,
        description="Organization identifiers should be taken from an existing organization identifier list. The scheme field is used to indicate the list or register from which the identifier is taken. This value should be taken from the [Organization Identifier Scheme](https://standard.open-contracting.org/1.1/en/schema/codelists/#organization-identifier-scheme) codelist.",
        title="Scheme",
    )
    id: Optional[Union[str, int]] = Field(
        None,
        description="The identifier of the organization in the selected scheme.",
        title="ID",
    )
    legalName: Optional[str] = Field(
        None,
        description="The legally registered name of the organization.",
        title="Legal Name",
    )
    # TODO This needs to implemented; maybe GLEIF can be used?
    # uri: Optional[AnyUrl] = Field(
    #     None,
    #     description="A URI to identify the organization, such as those provided by [Open Corporates](https://opencorporates.com) or some other relevant URI provider. This is not for listing the website of the organization: that can be done through the URL field of the Organization contact point.",
    #     title="URI",
    # )

    def __hash__(self) -> int:
        return super().__hash__()


class Address(BaseModel):
    streetAddress: Optional[str] = Field(
        None,
        description="The street address. For example, 1600 Amphitheatre Pkwy.",
        title="Street address",
    )
    locality: Optional[str] = Field(
        None, description="The locality. For example, Mountain View.", title="Locality"
    )
    region: Optional[str] = Field(
        None, description="The region. For example, CA.", title="Region"
    )
    postalCode: Optional[str] = Field(
        None, description="The postal code. For example, 94043.", title="Postal code"
    )
    countryName: Optional[str] = Field(
        None,
        description="The country name. For example, United States.",
        title="Country name",
    )

    def __hash__(self) -> int:
        return super().__hash__()


class ContactPoint(BaseModel):
    name: Optional[str] = Field(
        None,
        description="The name of the contact person, department, or contact point, for correspondence relating to this contracting process.",
        title="Name",
    )
    email: Optional[str] = Field(
        None,
        description="The e-mail address of the contact point/person.",
        title="Email",
    )
    telephone: Optional[str] = Field(
        None,
        description="The telephone number of the contact point/person. This should include the international dialing code.",
        title="Telephone",
    )
    faxNumber: Optional[str] = Field(
        None,
        description="The fax number of the contact point/person. This should include the international dialing code.",
        title="Fax number",
    )
    # TODO Properly validate URL using AnyUrl
    url: Optional[str] = Field(
        None, description="A web address for the contact point/person.", title="URL"
    )

    def __hash__(self) -> int:
        return super().__hash__()


class Currency(Enum):
    ADP = "ADP"
    AED = "AED"
    AFA = "AFA"
    AFN = "AFN"
    ALK = "ALK"
    ALL = "ALL"
    AMD = "AMD"
    ANG = "ANG"
    AOA = "AOA"
    AOK = "AOK"
    AON = "AON"
    AOR = "AOR"
    ARA = "ARA"
    ARP = "ARP"
    ARS = "ARS"
    ARY = "ARY"
    ATS = "ATS"
    AUD = "AUD"
    AWG = "AWG"
    AYM = "AYM"
    AZM = "AZM"
    AZN = "AZN"
    BAD = "BAD"
    BAM = "BAM"
    BBD = "BBD"
    BDT = "BDT"
    BEC = "BEC"
    BEF = "BEF"
    BEL = "BEL"
    BGJ = "BGJ"
    BGK = "BGK"
    BGL = "BGL"
    BGN = "BGN"
    BHD = "BHD"
    BIF = "BIF"
    BMD = "BMD"
    BND = "BND"
    BOB = "BOB"
    BOP = "BOP"
    BOV = "BOV"
    BRB = "BRB"
    BRC = "BRC"
    BRE = "BRE"
    BRL = "BRL"
    BRN = "BRN"
    BRR = "BRR"
    BSD = "BSD"
    BTN = "BTN"
    BUK = "BUK"
    BWP = "BWP"
    BYB = "BYB"
    BYN = "BYN"
    BYR = "BYR"
    BZD = "BZD"
    CAD = "CAD"
    CDF = "CDF"
    CHC = "CHC"
    CHE = "CHE"
    CHF = "CHF"
    CHW = "CHW"
    CLF = "CLF"
    CLP = "CLP"
    CNY = "CNY"
    COP = "COP"
    COU = "COU"
    CRC = "CRC"
    CSD = "CSD"
    CSJ = "CSJ"
    CSK = "CSK"
    CUC = "CUC"
    CUP = "CUP"
    CVE = "CVE"
    CYP = "CYP"
    CZK = "CZK"
    DDM = "DDM"
    DEM = "DEM"
    DJF = "DJF"
    DKK = "DKK"
    DOP = "DOP"
    DZD = "DZD"
    ECS = "ECS"
    ECV = "ECV"
    EEK = "EEK"
    EGP = "EGP"
    ERN = "ERN"
    ESA = "ESA"
    ESB = "ESB"
    ESP = "ESP"
    ETB = "ETB"
    EUR = "EUR"
    FIM = "FIM"
    FJD = "FJD"
    FKP = "FKP"
    FRF = "FRF"
    GBP = "GBP"
    GEK = "GEK"
    GEL = "GEL"
    GHC = "GHC"
    GHP = "GHP"
    GHS = "GHS"
    GIP = "GIP"
    GMD = "GMD"
    GNE = "GNE"
    GNF = "GNF"
    GNS = "GNS"
    GQE = "GQE"
    GRD = "GRD"
    GTQ = "GTQ"
    GWE = "GWE"
    GWP = "GWP"
    GYD = "GYD"
    HKD = "HKD"
    HNL = "HNL"
    HRD = "HRD"
    HRK = "HRK"
    HTG = "HTG"
    HUF = "HUF"
    IDR = "IDR"
    IEP = "IEP"
    ILP = "ILP"
    ILR = "ILR"
    ILS = "ILS"
    INR = "INR"
    IQD = "IQD"
    IRR = "IRR"
    ISJ = "ISJ"
    ISK = "ISK"
    ITL = "ITL"
    JMD = "JMD"
    JOD = "JOD"
    JPY = "JPY"
    KES = "KES"
    KGS = "KGS"
    KHR = "KHR"
    KMF = "KMF"
    KPW = "KPW"
    KRW = "KRW"
    KWD = "KWD"
    KYD = "KYD"
    KZT = "KZT"
    LAJ = "LAJ"
    LAK = "LAK"
    LBP = "LBP"
    LKR = "LKR"
    LRD = "LRD"
    LSL = "LSL"
    LSM = "LSM"
    LTL = "LTL"
    LTT = "LTT"
    LUC = "LUC"
    LUF = "LUF"
    LUL = "LUL"
    LVL = "LVL"
    LVR = "LVR"
    LYD = "LYD"
    MAD = "MAD"
    MDL = "MDL"
    MGA = "MGA"
    MGF = "MGF"
    MKD = "MKD"
    MLF = "MLF"
    MMK = "MMK"
    MNT = "MNT"
    MOP = "MOP"
    MRO = "MRO"
    MRU = "MRU"
    MTL = "MTL"
    MTP = "MTP"
    MUR = "MUR"
    MVQ = "MVQ"
    MVR = "MVR"
    MWK = "MWK"
    MXN = "MXN"
    MXP = "MXP"
    MXV = "MXV"
    MYR = "MYR"
    MZE = "MZE"
    MZM = "MZM"
    MZN = "MZN"
    NAD = "NAD"
    NGN = "NGN"
    NIC = "NIC"
    NIO = "NIO"
    NLG = "NLG"
    NOK = "NOK"
    NPR = "NPR"
    NZD = "NZD"
    OMR = "OMR"
    PAB = "PAB"
    PEH = "PEH"
    PEI = "PEI"
    PEN = "PEN"
    PES = "PES"
    PGK = "PGK"
    PHP = "PHP"
    PKR = "PKR"
    PLN = "PLN"
    PLZ = "PLZ"
    PTE = "PTE"
    PYG = "PYG"
    QAR = "QAR"
    RHD = "RHD"
    ROK = "ROK"
    ROL = "ROL"
    RON = "RON"
    RSD = "RSD"
    RUB = "RUB"
    RUR = "RUR"
    RWF = "RWF"
    SAR = "SAR"
    SBD = "SBD"
    SCR = "SCR"
    SDD = "SDD"
    SDG = "SDG"
    SDP = "SDP"
    SEK = "SEK"
    SGD = "SGD"
    SHP = "SHP"
    SIT = "SIT"
    SKK = "SKK"
    SLL = "SLL"
    SOS = "SOS"
    SRD = "SRD"
    SRG = "SRG"
    SSP = "SSP"
    STD = "STD"
    STN = "STN"
    SUR = "SUR"
    SVC = "SVC"
    SYP = "SYP"
    SZL = "SZL"
    THB = "THB"
    TJR = "TJR"
    TJS = "TJS"
    TMM = "TMM"
    TMT = "TMT"
    TND = "TND"
    TOP = "TOP"
    TPE = "TPE"
    TRL = "TRL"
    TRY = "TRY"
    TTD = "TTD"
    TWD = "TWD"
    TZS = "TZS"
    UAH = "UAH"
    UAK = "UAK"
    UGS = "UGS"
    UGW = "UGW"
    UGX = "UGX"
    USD = "USD"
    USN = "USN"
    USS = "USS"
    UYI = "UYI"
    UYN = "UYN"
    UYP = "UYP"
    UYU = "UYU"
    UYW = "UYW"
    UZS = "UZS"
    VEB = "VEB"
    VEF = "VEF"
    VES = "VES"
    VNC = "VNC"
    VND = "VND"
    VUV = "VUV"
    WST = "WST"
    XAF = "XAF"
    XAG = "XAG"
    XAU = "XAU"
    XBA = "XBA"
    XBB = "XBB"
    XBC = "XBC"
    XBD = "XBD"
    XCD = "XCD"
    XDR = "XDR"
    XEU = "XEU"
    XFO = "XFO"
    XFU = "XFU"
    XOF = "XOF"
    XPD = "XPD"
    XPF = "XPF"
    XPT = "XPT"
    XRE = "XRE"
    XSU = "XSU"
    XTS = "XTS"
    XUA = "XUA"
    XXX = "XXX"
    YDD = "YDD"
    YER = "YER"
    YUD = "YUD"
    YUM = "YUM"
    YUN = "YUN"
    ZAL = "ZAL"
    ZAR = "ZAR"
    ZMK = "ZMK"
    ZMW = "ZMW"
    ZRN = "ZRN"
    ZRZ = "ZRZ"
    ZWC = "ZWC"
    ZWD = "ZWD"
    ZWL = "ZWL"
    ZWN = "ZWN"
    ZWR = "ZWR"
    NoneType_None = None


class Value(BaseModel):
    amount: Optional[float] = Field(
        None, description="Amount as a number.", title="Amount"
    )
    amountEur: Optional[float] = Field(
        None, description="Amount in Euro if the default amount is in different currency", title="Amount in euro"
    )
    currency: Optional[Currency] = Field(
        None,
        description="The currency of the amount, from the closed [currency](https://standard.open-contracting.org/1.1/en/schema/codelists/#currency) codelist.",
        title="Currency",
    )


class Period(BaseModel):
    startDate: Optional[datetime] = Field(
        None,
        description="The start date for the period. When known, a precise start date must be provided.",
        title="Start date",
    )
    endDate: Optional[datetime] = Field(
        None,
        description="The end date for the period. When known, a precise end date must be provided.",
        title="End date",
    )
    maxExtentDate: Optional[datetime] = Field(
        None,
        description="The period cannot be extended beyond this date. This field can be used to express the maximum available date for extension or renewal of this period.",
        title="Maximum extent",
    )
    durationInDays: Optional[int] = Field(
        None,
        description="The maximum duration of this period in days. A user interface can collect or display this data in months or years as appropriate, and then convert it into days when storing this field. This field can be used when exact dates are not known. If a startDate and endDate are set, this field, if used, should be equal to the difference between startDate and endDate. Otherwise, if a startDate and maxExtentDate are set, this field, if used, should be equal to the difference between startDate and maxExtentDate.",
        title="Duration (days)",
    )


class RelatedProcess(BaseModel):
    id: Optional[str] = Field(
        None,
        description="A local identifier for this relationship, unique within this array.",
        title="Relationship ID",
    )
    relationship: Optional[List[str]] = Field(
        None,
        description="The type of relationship, using the open [relatedProcess](https://standard.open-contracting.org/1.1/en/schema/codelists/#related-process) codelist.",
        title="Relationship",
    )
    title: Optional[str] = Field(
        None,
        description="The title of the related process, where referencing an open contracting process, this field should match the tender/title field in the related process.",
        title="Related process title",
    )
    scheme: Optional[str] = Field(
        None,
        description="The identification scheme used by this cross-reference, using the open [relatedProcessScheme](https://standard.open-contracting.org/1.1/en/schema/codelists/#related-process-scheme) codelist.",
        title="Scheme",
    )
    identifier: Optional[str] = Field(
        None,
        description="The identifier of the related process. If the scheme is 'ocid', this must be an Open Contracting ID (ocid).",
        title="Identifier",
    )
    uri: Optional[AnyUrl] = Field(
        None,
        description="A URI pointing to a machine-readable document, release or record package containing the identified related process.",
        title="Related process URI",
    )

    def __hash__(self):
        return hash(self.id)


class Milestone(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="A local identifier for this milestone, unique within this block. This field is used to keep track of multiple revisions of a milestone through the compilation from release to record mechanism.",
        title="ID",
    )
    title: Optional[str] = Field(None, description="Milestone title", title="Title")
    type: Optional[str] = Field(
        None,
        description="The nature of the milestone, using the open [milestoneType](https://standard.open-contracting.org/1.1/en/schema/codelists/#milestone-type) codelist.",
        title="Milestone type",
    )
    description: Optional[str] = Field(
        None, description="A description of the milestone.", title="Description"
    )
    code: Optional[str] = Field(
        None,
        description="Milestone codes can be used to track specific events that take place for a particular kind of contracting process. For example, a code of 'approvalLetter' can be used to allow applications to understand this milestone represents the date an approvalLetter is due or signed.",
        title="Milestone code",
    )
    dueDate: Optional[datetime] = Field(
        None, description="The date the milestone is due.", title="Due date"
    )
    dateMet: Optional[datetime] = Field(
        None, description="The date on which the milestone was met.", title="Date met"
    )
    dateModified: Optional[datetime] = Field(
        None,
        description="The date the milestone was last reviewed or modified and the status was altered or confirmed to still be correct.",
        title="Date modified",
    )
    status: Optional[MilestoneStatus] = Field(
        None,
        description="The status that was realized on the date provided in `dateModified`, from the closed [milestoneStatus](https://standard.open-contracting.org/1.1/en/schema/codelists/#milestone-status) codelist.",
        title="Status",
    )
    documents: Optional[Set[Document]] = Field(
        None,
        description="List of documents associated with this milestone (Deprecated in 1.1).",
        title="Documents",
    )


class Budget(BaseModel):
    id: Optional[Union[str, int]] = Field(
        None,
        description="An identifier for the budget line item which provides funds for this contracting process. This identifier should be possible to cross-reference against the provided data source.",
        title="ID",
    )
    description: Optional[str] = Field(
        None,
        description="A short free text description of the budget source. May be used to provide the title of the budget line, or the programme used to fund this project.",
        title="Budget Source",
    )
    amount: Optional[Value] = Field(
        None,
        description="The value reserved in the budget for this contracting process. A negative value indicates anticipated income to the budget as a result of this contracting process, rather than expenditure. Where the budget is drawn from multiple sources, the budget breakdown extension can be used.",
        title="Amount",
    )
    project: Optional[str] = Field(
        None,
        description="The name of the project through which this contracting process is funded (if applicable). Some organizations maintain a registry of projects, and the data should use the name by which the project is known in that registry. No translation option is offered for this string, as translated values can be provided in third-party data, linked from the data source above.",
        title="Project title",
    )
    projectID: Optional[Union[str, int]] = Field(
        None,
        description="An external identifier for the project that this contracting process forms part of, or is funded via (if applicable). Some organizations maintain a registry of projects, and the data should use the identifier from the relevant registry of projects.",
        title="Project identifier",
    )
    uri: Optional[AnyUrl] = Field(
        None,
        description="A URI pointing directly to a machine-readable record about the budget line-item or line-items that fund this contracting process. Information can be provided in a range of formats, including using IATI, the Open Fiscal Data Standard or any other standard which provides structured data on budget sources. Human readable documents can be included using the planning.documents block.",
        title="Linked budget information",
    )
    source: Optional[AnyUrl] = Field(
        None,
        description="(Deprecated in 1.1) Used to point either to a corresponding Budget Data Package, or to a machine or human-readable source where users can find further information on the budget line item identifiers, or project identifiers, provided here.",
        title="Data Source",
    )


class Organization(BaseModel):
    name: Optional[str] = Field(
        None,
        description="A common name for this organization or other participant in the contracting process. The identifier object provides a space for the formal legal name, and so this may either repeat that value, or may provide the common name by which this organization or entity is known. This field may also include details of the department or sub-unit involved in this contracting process.",
        title="Common name",
    )
    id: Optional[str] = Field(
        None,
        description="The ID used for cross-referencing to this party from other sections of the release. This field may be built with the following structure {identifier.scheme}-{identifier.id}(-{department-identifier}).",
        title="Entity ID",
    )
    identifier: Optional[Identifier] = Field(
        None,
        description="The primary identifier for this organization or participant. Identifiers that uniquely pick out a legal entity should be preferred. Consult the [organization identifier guidance](https://standard.open-contracting.org/1.1/en/schema/identifiers/) for the preferred scheme and identifier to use.",
        title="Primary identifier",
    )
    additionalIdentifiers: Optional[Set[Identifier]] = Field(
        None,
        description="A list of additional / supplemental identifiers for the organization or participant, using the [organization identifier guidance](https://standard.open-contracting.org/1.1/en/schema/identifiers/). This can be used to provide an internally used identifier for this organization in addition to the primary legal entity identifier.",
        title="Additional identifiers",
    )
    address: Optional[Address] = Field(
        None,
        description="An address. This may be the legally registered address of the organization, or may be a correspondence address for this particular contracting process.",
        title="Address",
    )
    contactPoint: Optional[ContactPoint] = Field(
        None,
        description="Contact details that can be used for this party.",
        title="Contact point",
    )
    roles: Optional[List[str]] = Field(
        None,
        description="The party's role(s) in the contracting process, using the open [partyRole](https://standard.open-contracting.org/1.1/en/schema/codelists/#party-role) codelist.",
        title="Party roles",
    )
    details: Optional[Dict[str, Any]] = Field(
        None,
        description="Additional classification information about parties can be provided using partyDetail extensions that define particular fields and classification schemes.",
        title="Details",
    )

    def __hash__(self):
        return hash(self.id)


class Unit(BaseModel):
    scheme: Optional[str] = Field(
        None,
        description="The list from which identifiers for units of measure are taken, using the open [unitClassificationScheme](https://standard.open-contracting.org/1.1/en/schema/codelists/#unit-classification-scheme) codelist. 'UNCEFACT' is recommended.",
        title="Scheme",
    )
    id: Optional[str] = Field(
        None,
        description="The identifier from the codelist referenced in the `scheme` field. Check the [unitClassificationScheme](https://standard.open-contracting.org/1.1/en/schema/codelists/#unit-classification-scheme) codelist for details of how to find and use identifiers from the scheme in use.",
        title="ID",
    )
    name: Optional[str] = Field(None, description="Name of the unit.", title="Name")
    value: Optional[Value] = Field(
        None, description="The monetary value of a single unit.", title="Value"
    )
    uri: Optional[AnyUrl] = Field(
        None,
        description="The machine-readable URI for the unit of measure, provided by the scheme.",
        title="URI",
    )


class Item(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="A local identifier to reference and merge the items by. Must be unique within a given array of items.",
        title="ID",
    )
    description: Optional[str] = Field(
        None,
        description="A description of the goods, services to be provided.",
        title="Description",
    )
    classification: Optional[Classification] = Field(
        None,
        description="The primary classification for the item.",
        title="Classification",
    )
    additionalClassifications: Optional[Set[Classification]] = Field(
        None,
        description="An array of additional classifications for the item.",
        title="Additional classifications",
    )
    quantity: Optional[float] = Field(
        None, description="The number of units to be provided.", title="Quantity"
    )
    unit: Optional[Unit] = Field(
        None,
        description="A description of the unit in which the supplies, services or works are provided (e.g. hours, kilograms) and the unit-price.",
        title="Unit",
    )


class Planning(BaseModel):
    rationale: Optional[str] = Field(
        None,
        description="The rationale for the procurement provided in free text. More detail can be provided in an attached document.",
        title="Rationale",
    )
    budget: Optional[Budget] = Field(
        None,
        description="Details of the budget that funds this contracting process.",
        title="Budget",
    )
    documents: Optional[List[Document]] = Field(
        None,
        description="A list of documents related to the planning process.",
        title="Documents",
    )
    milestones: Optional[List[Milestone]] = Field(
        None,
        description="A list of milestones associated with the planning stage.",
        title="Planning milestones",
    )


class Tender(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="An identifier for this tender process. This may be the same as the ocid, or may be an internal identifier for this tender.",
        title="Tender ID",
    )
    title: Optional[str] = Field(
        None,
        description="A title for this tender. This will often be used by applications as a headline to attract interest, and to help analysts understand the nature of this procurement.",
        title="Tender title",
    )
    description: Optional[str] = Field(
        None,
        description="A summary description of the tender. This complements any structured information provided using the items array. Descriptions should be short and easy to read. Avoid using ALL CAPS.",
        title="Tender description",
    )
    status: Optional[Status] = Field(
        None,
        description="The current status of the tender, from the closed [tenderStatus](https://standard.open-contracting.org/1.1/en/schema/codelists/#tender-status) codelist.",
        title="Tender status",
    )
    procuringEntity: Optional[Organization] = Field(
        None,
        description="The entity managing the procurement. This may be different from the buyer who pays for, or uses, the items being procured.",
        title="Procuring entity",
    )
    items: Optional[Set[Item]] = Field(
        None,
        description="The goods and services to be purchased, broken into line items wherever possible. Items should not be duplicated, but the quantity specified instead.",
        title="Items to be procured",
    )
    value: Optional[Value] = Field(
        None,
        description="The total upper estimated value of the procurement. A negative value indicates that the contracting process may involve payments from the supplier to the buyer (commonly used in concession contracts).",
        title="Value",
    )
    minValue: Optional[Value] = Field(
        None,
        description="The minimum estimated value of the procurement.  A negative value indicates that the contracting process may involve payments from the supplier to the buyer (commonly used in concession contracts).",
        title="Minimum value",
    )
    procurementMethod: Optional[ProcurementMethod] = Field(
        None,
        description="The procurement method, from the closed [method](https://standard.open-contracting.org/1.1/en/schema/codelists/#method) codelist.",
        title="Procurement method",
    )
    procurementMethodDetails: Optional[str] = Field(
        None,
        description="Additional detail on the procurement method used. This field can be used to provide the local name of the particular procurement method used.",
        title="Procurement method details",
    )
    procurementMethodRationale: Optional[str] = Field(
        None,
        description="Rationale for the chosen procurement method. This is especially important to provide a justification in the case of limited tenders or direct awards.",
        title="Procurement method rationale",
    )
    mainProcurementCategory: Optional[MainProcurementCategory] = Field(
        None,
        description="The primary category describing the main object of this contracting process, from the closed [procurementCategory](https://standard.open-contracting.org/1.1/en/schema/codelists/#procurement-category) codelist.",
        title="Main procurement category",
    )
    additionalProcurementCategories: Optional[List[str]] = Field(
        None,
        description="Any additional categories describing the objects of this contracting process, using the open [extendedProcurementCategory](https://standard.open-contracting.org/1.1/en/schema/codelists/#extended-procurement-category) codelist.",
        title="Additional procurement categories",
    )
    awardCriteria: Optional[str] = Field(
        None,
        description="The award criteria for the procurement, using the open [awardCriteria](https://standard.open-contracting.org/1.1/en/schema/codelists/#award-criteria) codelist.",
        title="Award criteria",
    )
    awardCriteriaDetails: Optional[str] = Field(
        None,
        description="Any detailed or further information on the award or selection criteria.",
        title="Award criteria details",
    )
    submissionMethod: Optional[List[str]] = Field(
        None,
        description="The methods by which bids are submitted, using the open [submissionMethod](https://standard.open-contracting.org/1.1/en/schema/codelists/#submission-method) codelist.",
        title="Submission method",
    )
    submissionMethodDetails: Optional[str] = Field(
        None,
        description="Any detailed or further information on the submission method. This can include the address, e-mail address or online service to which bids are submitted, and any special requirements to be followed for submissions.",
        title="Submission method details",
    )
    tenderPeriod: Optional[Period] = Field(
        None,
        description="The period when the tender is open for submissions. The end date is the closing date for tender submissions.",
        title="Tender period",
    )
    enquiryPeriod: Optional[Period] = Field(
        None,
        description="The period during which potential bidders may submit questions and requests for clarification to the entity managing procurement. Details of how to submit enquiries should be provided in attached notices, or in submissionMethodDetails. Structured dates for when responses to questions will be made can be provided using tender milestones.",
        title="Enquiry period",
    )
    hasEnquiries: Optional[bool] = Field(
        None,
        description="A true/false field to indicate whether any enquiries were received during the tender process. Structured information on enquiries that were received, and responses to them, can be provided using the enquiries extension.",
        title="Has enquiries?",
    )
    eligibilityCriteria: Optional[str] = Field(
        None,
        description="A description of any eligibility criteria for potential suppliers.",
        title="Eligibility criteria",
    )
    awardPeriod: Optional[Period] = Field(
        None,
        description="The period for decision making regarding the contract award. The end date should be the date on which an award decision is due to be finalized. The start date may be used to indicate the start of an evaluation period.",
        title="Evaluation and award period",
    )
    contractPeriod: Optional[Period] = Field(
        None,
        description="The period over which the contract is estimated or required to be active. If the tender does not specify explicit dates, the duration field may be used.",
        title="Contract period",
    )
    numberOfTenderers: Optional[int] = Field(
        None,
        description="The number of parties who submit a bid.",
        title="Number of tenderers",
    )
    tenderers: Optional[Set[Organization]] = Field(
        None,
        description="All parties who submit a bid on a tender. More detailed information on bids and the bidding organization can be provided using the bid extension.",
        title="Tenderers",
    )
    documents: Optional[List[Document]] = Field(
        None,
        description="All documents and attachments related to the tender, including any notices. See the [documentType](https://standard.open-contracting.org/1.1/en/schema/codelists/#document-type) codelist for details of potential documents to include. Common documents include official legal notices of tender, technical specifications, evaluation criteria, and, as a tender process progresses, clarifications and replies to queries.",
        title="Documents",
    )
    milestones: Optional[List[Milestone]] = Field(
        None,
        description="A list of milestones associated with the tender.",
        title="Milestones",
    )
    amendments: Optional[List[Amendment]] = Field(
        None,
        description="A tender amendment is a formal change to the tender, and generally involves the publication of a new tender notice/release. The rationale and a description of the changes made can be provided here.",
        title="Amendments",
    )
    amendment: Optional[Amendment] = Field(
        None,
        description="The use of individual amendment objects has been deprecated. From OCDS 1.1 information should be provided in the amendments array.",
        title="Amendment",
    )
    # Extensions
    crossBorderLaw: Optional[str] = Field(
        None, description="Cross border law", title="Cross border law"
    )
    classification: Optional[Classification] = Field(
        None, description="Tender classification", title="Classification"
    )

    def __hash__(self):
        return hash(self.id)


class Award(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="The identifier for this award. It must be unique and must not change within the Open Contracting Process it is part of (defined by a single ocid). See the [identifier guidance](https://standard.open-contracting.org/1.1/en/schema/identifiers/) for further details.",
        title="Award ID",
    )
    title: Optional[str] = Field(None, description="Award title", title="Title")
    description: Optional[str] = Field(
        None, description="Award description", title="Description"
    )
    status: Optional[AwardStatus] = Field(
        None,
        description="The current status of the award, from the closed [awardStatus](https://standard.open-contracting.org/1.1/en/schema/codelists/#award-status) codelist.",
        title="Award status",
    )
    date: Optional[datetime] = Field(
        None,
        description="The date of the contract award. This is usually the date on which a decision to award was made.",
        title="Award date",
    )
    value: Optional[Value] = Field(
        None,
        description="The total value of this award. In the case of a framework contract this may be the total estimated lifetime value, or maximum value, of the agreement. There may be more than one award per procurement. A negative value indicates that the award may involve payments from the supplier to the buyer (commonly used in concession contracts).",
        title="Value",
    )
    suppliers: Optional[Set[Organization]] = Field(
        None,
        description="The suppliers awarded this award. If different suppliers have been awarded different items or values, these should be split into separate award blocks.",
        title="Suppliers",
    )
    items: Optional[Set[Item]] = Field(
        None,
        description="The goods and services awarded in this award, broken into line items wherever possible. Items should not be duplicated, but the quantity specified instead.",
        min_items=1,
        title="Items awarded",
    )
    contractPeriod: Optional[Period] = Field(
        None,
        description="The period for which the contract has been awarded.",
        title="Contract period",
    )
    documents: Optional[Set[Document]] = Field(
        None,
        description="All documents and attachments related to the award, including any notices.",
        title="Documents",
    )
    amendments: Optional[List[Amendment]] = Field(
        None,
        description="An award amendment is a formal change to the details of the award, and generally involves the publication of a new award notice/release. The rationale and a description of the changes made can be provided here.",
        title="Amendments",
    )
    amendment: Optional[Amendment] = Field(
        None,
        description="The use of individual amendment objects has been deprecated. From OCDS 1.1 information should be provided in the amendments array.",
        title="Amendment",
    )

    def __hash__(self):
        return hash(self.id)


class Transaction(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="A unique identifier for this transaction. This identifier should be possible to cross-reference against the provided data source. For IATI this is the transaction reference.",
        title="ID",
    )
    source: Optional[AnyUrl] = Field(
        None,
        description="Used to point either to a corresponding Fiscal Data Package, IATI file, or machine or human-readable source where users can find further information on the budget line item identifiers, or project identifiers, provided here.",
        title="Data source",
    )
    date: Optional[datetime] = Field(
        None, description="The date of the transaction", title="Date"
    )
    value: Optional[Value] = Field(
        None, description="The value of the transaction.", title="Value"
    )
    payer: Optional[Organization] = Field(
        None,
        description="An organization reference for the organization from which the funds in this transaction originate.",
        title="Payer",
    )
    payee: Optional[Organization] = Field(
        None,
        description="An organization reference for the organization which receives the funds in this transaction.",
        title="Payee",
    )
    uri: Optional[AnyUrl] = Field(
        None,
        description="A URI pointing directly to a machine-readable record about this spending transaction.",
        title="Linked spending information",
    )
    amount: Optional[Value] = Field(
        None,
        description="(Deprecated in 1.1. Use transaction.value instead) The value of the transaction. A negative value indicates a refund or correction.",
        title="Amount",
    )
    providerOrganization: Optional[Identifier] = Field(
        None,
        description="(Deprecated in 1.1. Use transaction.payer instead.) The Organization Identifier for the organization from which the funds in this transaction originate. Expressed following the Organizational Identifier standard - consult the documentation and the codelist.",
        title="Provider organization",
    )
    receiverOrganization: Optional[Identifier] = Field(
        None,
        description="(Deprecated in 1.1. Use transaction.payee instead). The Organization Identifier for the organization which receives the funds in this transaction. Expressed following the Organizational Identifier standard - consult the documentation and the codelist.",
        title="Receiver organization",
    )


class Implementation(BaseModel):
    transactions: Optional[Set[Transaction]] = Field(
        None,
        description="A list of the spending transactions made against this contract",
        title="Transactions",
    )
    milestones: Optional[Set[Milestone]] = Field(
        None,
        description="As milestones are completed, the milestone's status and dates should be updated.",
        title="Milestones",
    )
    documents: Optional[Set[Document]] = Field(
        None,
        description="Documents and reports that are part of the implementation phase e.g. audit and evaluation reports.",
        title="Documents",
    )


class Contract(BaseModel):
    id: Union[constr(min_length=1), int] = Field(
        ...,
        description="The identifier for this contract. It must be unique and must not change within the Open Contracting Process it is part of (defined by a single ocid). See the [identifier guidance](https://standard.open-contracting.org/1.1/en/schema/identifiers/) for further details.",
        title="Contract ID",
    )
    awardID: Union[constr(min_length=1), int] = Field(
        ...,
        description="The award.id against which this contract is being issued.",
        title="Award ID",
    )
    title: Optional[str] = Field(
        None, description="Contract title", title="Contract title"
    )
    description: Optional[str] = Field(
        None, description="Contract description", title="Contract description"
    )
    status: Optional[ContractStatus] = Field(
        None,
        description="The current status of the contract, from the closed [contractStatus](https://standard.open-contracting.org/1.1/en/schema/codelists/#contract-status) codelist.",
        title="Contract status",
    )
    period: Optional[Period] = Field(
        None, description="The start and end date for the contract.", title="Period"
    )
    value: Optional[Value] = Field(
        None,
        description="The total value of this contract. A negative value indicates that the contract will involve payments from the supplier to the buyer (commonly used in concession contracts).",
        title="Value",
    )
    items: Optional[Set[Item]] = Field(
        None,
        description="The goods, services, and any intangible outcomes in this contract. Note: If the items are the same as the award do not repeat.",
        min_items=1,
        title="Items contracted",
    )
    dateSigned: Optional[datetime] = Field(
        None,
        description="The date the contract was signed. In the case of multiple signatures, the date of the last signature.",
        title="Date signed",
    )
    documents: Optional[Set[Document]] = Field(
        None,
        description="All documents and attachments related to the contract, including any notices.",
        title="Documents",
    )
    implementation: Optional[Implementation] = Field(
        None,
        description="Information related to the implementation of the contract in accordance with the obligations laid out therein.",
        title="Implementation",
    )
    relatedProcesses: Optional[Set[RelatedProcess]] = Field(
        None,
        description="The details of related processes: for example, if this process is followed by one or more contracting processes, represented under a separate open contracting identifier (ocid). This is commonly used to refer to subcontracts and to renewal or replacement processes for this contract.",
        title="Related processes",
    )
    milestones: Optional[List[Milestone]] = Field(
        None,
        description="A list of milestones associated with the finalization of this contract.",
        title="Contract milestones",
    )
    amendments: Optional[List[Amendment]] = Field(
        None,
        description="A contract amendment is a formal change to, or extension of, a contract, and generally involves the publication of a new contract notice/release, or some other documents detailing the change. The rationale and a description of the changes made can be provided here.",
        title="Amendments",
    )
    amendment: Optional[Amendment] = Field(
        None,
        description="The use of individual amendment objects has been deprecated. From OCDS 1.1 information should be provided in the amendments array.",
        title="Amendment",
    )

    def __hash__(self):
        return hash(self.id)


class Release(BaseModel):
    ocid: constr(min_length=1) = Field(
        None,
        description="A globally unique identifier for this Open Contracting Process. Composed of an ocid prefix and an identifier for the contracting process. For more information see the [Open Contracting Identifier guidance](https://standard.open-contracting.org/1.1/en/schema/identifiers/)",
        title="Open Contracting ID",
    )
    id: constr(min_length=1) = Field(
        None,
        description="An identifier for this particular release of information. A release identifier must be unique within the scope of its related contracting process (defined by a common ocid). A release identifier must not contain the # character.",
        title="Release ID",
    )
    date: datetime = Field(
        None,
        description="The date on which the information contained in the release was first recorded in, or published by, any system.",
        title="Release Date",
    )
    tag: List[TagEnum] = Field(
        None,
        description="One or more values from the closed [releaseTag](https://standard.open-contracting.org/1.1/en/schema/codelists/#release-tag) codelist. Tags can be used to filter releases and to understand the kind of information that releases might contain.",
        min_items=1,
        title="Release Tag",
    )
    initiationType: InitiationType = Field(
        None,
        description="The type of initiation process used for this contract, from the closed [initiationType](https://standard.open-contracting.org/1.1/en/schema/codelists/#initiation-type) codelist.",
        title="Initiation type",
    )
    parties: Optional[Set[Organization]] = Field(
        None,
        description="Information on the parties (organizations, economic operators and other participants) who are involved in the contracting process and their roles, e.g. buyer, procuring entity, supplier etc. Organization references elsewhere in the schema are used to refer back to this entries in this list.",
        title="Parties",
    )
    buyer: Optional[Organization] = Field(
        None,
        description="A buyer is an entity whose budget will be used to pay for goods, works or services related to a contract. This may be different from the procuring entity who may be specified in the tender data.",
        title="Buyer",
    )
    planning: Optional[Planning] = Field(
        None,
        description="Information from the planning phase of the contracting process. This includes information related to the process of deciding what to contract, when and how.",
        title="Planning",
    )
    tender: Optional[Tender] = Field(
        None,
        description="The activities undertaken in order to enter into a contract.",
        title="Tender",
    )
    awards: Optional[Set[Award]] = Field(
        None,
        description="Information from the award phase of the contracting process. There can be more than one award per contracting process e.g. because the contract is split among different providers, or because it is a standing offer.",
        title="Awards",
    )
    contracts: Optional[Set[Contract]] = Field(
        None,
        description="Information from the contract creation phase of the procurement process.",
        title="Contracts",
    )
    language: Optional[str] = Field(
        "en",
        description="The default language of the data using either two-letter [ISO639-1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes), or extended [BCP47 language tags](https://www.w3.org/International/articles/language-tags/). The use of lowercase two-letter codes from [ISO639-1](https://en.wikipedia.org/wiki/List_of_ISO_639-1_codes) is recommended.",
        title="Release language",
    )
    relatedProcesses: Optional[Set[RelatedProcess]] = Field(
        None,
        description="The details of related processes: for example, if this process follows on from one or more other processes, represented under a separate open contracting identifier (ocid). This is commonly used to relate mini-competitions to their parent frameworks or individual tenders to a broader planning process.",
        title="Related processes",
    )
    # Link to TED
    tedURL: Optional[str]

    def __hash__(self):
        return hash(self.id)



import argparse
import logging
import multiprocessing
import os
import shutil
import tarfile
from datetime import datetime
from urllib.request import urlretrieve

import luigi
import orjson
import orjsonl
import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from dateutil import rrule
from dateutil.relativedelta import relativedelta
from kuzu import Connection, Database
from luigi import LocalTarget, Parameter, Task
from luigi.execution_summary import LuigiRunResult, LuigiStatusCode

from tedective_etl.currency.update import fetch_exchange_rates
from tedective_etl.dedupe import dedupe_organizations
from tedective_etl.graph import (create_import_files, create_kuzudb_schema,
                                 import_to_kuzu)
from tedective_etl.search import create_index
from tedective_etl.ted_to_ocds.process import (get_ted_xml_files,
                                               process_ted_notice)
from tedective_etl.ted_to_ocds.transform import implemented_forms
from tedective_etl.utils import prune_list


class DownloadCurrencyRates(Task):
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)
    output_dir = Parameter(default=None)

    def run(self):
        fetch_exchange_rates(self.first_month, self.last_month)

        with self.output().open("w") as f:
            f.write("Fetching exchange rates was completed successfully.")

    def output(self):
        return LocalTarget(os.path.join(self.output_dir, "download_currency_rates_complete.txt"))


class DownloadTEDNotices(Task):
    in_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)

    def run(self):
        os.makedirs(self.in_dir, exist_ok=True)
        #fetch_exchange_rates(self.first_month, self.last_month)

        start_date = datetime(
            int(self.first_month.split("-")[0]),
            int(self.first_month.split("-")[1]),
            1,
        )
        end_date = datetime(
            int(self.last_month.split("-")[0]),
            int(self.last_month.split("-")[1]),
            1,
        )

        for dt in rrule.rrule(rrule.MONTHLY, dtstart=start_date, until=end_date):
            year_month = dt.strftime("%Y-%m")
            if not os.path.isdir(f"{self.in_dir}/{year_month}"):
                if not os.path.isfile(f"{self.in_dir}/{year_month}.tar.gz"):
                    logging.info(f"Downloading TED XMLs for {year_month}...")
                    url = f"https://ted.europa.eu/packages/monthly/{year_month}"
                    dst = f"{self.in_dir}/{year_month}.tar.gz"
                    urlretrieve(url, dst)

                logging.info(f"Extracting TED XMLs for {year_month}...")
                f_to_extract = tarfile.open(f"{self.in_dir}/{year_month}.tar.gz")
                f_to_extract.extractall(f"{self.in_dir}/{year_month}")
                os.remove(f"{self.in_dir}/{year_month}.tar.gz")

                if year_month == self.last_month:
                    with open(
                        f"{self.in_dir}/{self.last_month}/complete.txt", "w"
                    ) as f:
                        f.write(
                            f"Download and extraction until {self.last_month} complete."
                        )

            else:
                logging.info(
                    f"Already downloaded and extracted TED XML for {year_month}."
                )

    def output(self):
        return LocalTarget(os.path.join(self.in_dir, self.last_month))


class ProcessTEDNotices(Task):
    in_dir = Parameter(default=None)
    output_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)

    def requires(self):
        return (
            DownloadTEDNotices(
                in_dir=self.in_dir,
                first_month=self.first_month,
                last_month=self.last_month,
            ),
            DownloadCurrencyRates(
                first_month=self.first_month,
                last_month=self.last_month,
                output_dir=self.output_dir,
            ),
        )

    def run(self):
        # Delete the output_dir if it exists
        if os.path.exists(self.output_dir):
            shutil.rmtree(self.output_dir)

        # Ensure the destination directory exists
        os.makedirs(self.output_dir, exist_ok=True)

        # Get a list of the downloaded TED XML files
        ted_xml_files = get_ted_xml_files(
            in_dir=self.in_dir,
            first_month=self.first_month,
            last_month=self.last_month,
        )
        seen_org_ids = set()

        # Initialize counters for covered and uncovered XMLs
        num_uncovered = 0
        num_covered = 0

        # Set the previous month to the first_month parameter
        previous_month = self.first_month

        # Iterate over the XML files (they are grouped by day)
        for day, xml_files in ted_xml_files.items():
            # Get number of CPUs
            num_cpus = multiprocessing.cpu_count()
            # Process each XML file in parallel
            with multiprocessing.Pool(num_cpus - 4) as pool:
                results = pool.map(
                    process_ted_notice,
                    [(xml_file) for xml_file in xml_files],
                )

            # Initialize lists to store relevant data
            releases = []
            organizations = []

            # Split results into the appropriate lists
            for r in results:
                if r["form_type"] not in implemented_forms:
                    # Add to the number of uncovered XMLs
                    num_uncovered += 1
                else:
                    # Get the relevant data
                    releases += r["releases"]
                    organizations += r["organizations"]

                    # Add to the number of covered XMLs
                    num_covered += 1

            # Prune the organizations
            pruned_organizations = prune_list(organizations, seen_org_ids)
            seen_org_ids = pruned_organizations[1]

            # Write organizations to disk (append-only)
            with open(f"{self.output_dir}/organizations.jsonl", "a") as f:
                for o in pruned_organizations[0]:
                    f.write(
                        orjson.dumps(
                            o.model_dump(mode="json", exclude_none=False)
                        ).decode("utf-8")
                        + "\n"
                    )

            # Prune releases
            pruned_releases = list(set(releases))

            # Get the month as YYYY-MM
            month = day[:6]
            month = month[:4] + "-" + month[4:]

            # Write releases to disk (append-only)
            with open(f"{self.output_dir}/{month}_ocds_releases.json", "a") as f:
                for r in pruned_releases:
                    f.write(
                        orjson.dumps(
                            r.model_dump(mode="json", exclude_none=False)
                        ).decode("utf-8")
                        + "\n"
                    )

            # Calculate the coverage ratio
            coverage = num_covered / (num_covered + num_uncovered) * 100

            # This logs when the month changes, e.g. from 2017-01 to 2017-02
            if month != previous_month:
                logging.info(
                    f"Processed TED XMLs for {previous_month}. Overall coverage: {coverage:.2f}%"
                )
                previous_month = month

        # This logs only once at the end
        if month == self.last_month:
            logging.info(
                f"Processed TED XMLs for {self.last_month}. Overall coverage: {coverage:.2f}%"
            )
            previous_month = month

        # Create a dataframe to hold all the OCDS releases
        logging.info("Creating OCDS dataframe...")
        ocds_df = pd.DataFrame()

        # Transform the JSONL files to OCDS 1.1.5
        logging.info("Transforming JSON files to OCDS 1.1.5...")
        for filename in os.listdir(self.output_dir):
            if filename.endswith(".json") and not filename.startswith("organizations"):
                # Load the JSONL files
                filepath = os.path.join(self.output_dir, filename)
                data = orjsonl.load(filepath)
                # Only transform the OCDS releases, not the organizations
                # There will be one file for each month
                new_data = {
                    "version": "1.1.5",
                    "publishedDate": datetime.now(),
                    "releases": data,
                    "publisher": "TEDective Team",
                    "license": "ODbL-1.0",
                }
                # Append to dataframe
                tmp_df = pd.DataFrame(data)
                ocds_df = pd.concat([ocds_df, tmp_df])

                # Write to file
                with open(filepath, "w") as f:
                    f.write(orjson.dumps(new_data).decode("utf-8"))

        # Write ocds_df to parquet file
        logging.info("Write OCDS dataframe to parquet file")
        ocds_table = pa.Table.from_pandas(ocds_df, preserve_index=False)
        pq.write_table(ocds_table, os.path.join(self.output_dir, "ocds_data.parquet"))

    def output(self):
        return LocalTarget(
            os.path.join(self.output_dir, f"{self.last_month}_ocds_releases.json")
        )


class BuildSearchIndex(Task):
    meilisearch_url = Parameter(default=None)
    sleep_time = Parameter(default=3)
    in_dir = Parameter(default=None)
    output_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)

    def requires(self):
        return ProcessTEDNotices(
            in_dir=self.in_dir,
            output_dir=self.output_dir,
            first_month=self.first_month,
            last_month=self.last_month,
        )

    def run(self):
        # Create the search index
        create_index(
            meilisearch_url=self.meilisearch_url,
            output_path=self.output_dir,
            sleep_time=self.sleep_time,
        )
        # Write a file to indicate that the search index has been built
        with self.output().open("w") as f:
            f.write("Search index built successfully.")

    def output(self):
        return LocalTarget(os.path.join(self.output_dir, "search_index_complete.txt"))


class CreateKuzuDBImportFiles(Task):
    in_dir = Parameter(default=None)
    output_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)

    def requires(self):
        return ProcessTEDNotices(
            in_dir=self.in_dir,
            output_dir=self.output_dir,
            first_month=self.first_month,
            last_month=self.last_month,
        )

    def run(self):
        create_import_files(
            path_parquet_data=f"{self.output_dir}/ocds_data.parquet",
            path_organization_json=f"{self.output_dir}/organizations.jsonl",
            path_output=self.output_dir,
        )

    def output(self):
        return LocalTarget(os.path.join(self.output_dir, "edges_hasSuppliers.parquet"))


class DedupeOrganizations(Task):
    in_dir = Parameter(default=None)
    output_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)

    def requires(self):
        return ProcessTEDNotices(
            in_dir=self.in_dir,
            output_dir=self.output_dir,
            first_month=self.first_month,
            last_month=self.last_month,
        )

    def run(self):
        dedupe_organizations(
            path_organization_json=f"{self.output_dir}/organizations.jsonl",
            path_output=self.output_dir,
        )

    def output(self):
        return LocalTarget(os.path.join(self.output_dir, "edges_isSameAs.parquet"))


class LoadKuzuDB(Task):
    in_dir = Parameter(default=None)
    output_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)
    graph_dir = Parameter(default=None)

    def requires(self):
        return (
            ProcessTEDNotices(
                in_dir=self.in_dir,
                output_dir=self.output_dir,
                first_month=self.first_month,
                last_month=self.last_month,
            ),
            CreateKuzuDBImportFiles(
                in_dir=self.in_dir,
                output_dir=self.output_dir,
                first_month=self.first_month,
                last_month=self.last_month,
            ),
            DedupeOrganizations(
                in_dir=self.in_dir,
                output_dir=self.output_dir,
                first_month=self.first_month,
                last_month=self.last_month,
            ),
        )

    def run(self):
        # Remove graph at graph_dir if it already exists using shutil.rmtree
        if os.path.exists(self.graph_dir):
            shutil.rmtree(self.graph_dir)

        # Create the database and connection
        db = Database(self.graph_dir)
        conn = Connection(db)

        # Create the schema
        create_kuzudb_schema(conn)

        # Import the data
        import_to_kuzu(self.output_dir, conn)

        # Write a file to indicate that the data has been loaded
        with self.output().open("w") as f:
            f.write("Data loaded into KuzuDB successfully.")

    def output(self):
        return LocalTarget(os.path.join(self.output_dir, "load_kuzudb_complete.txt"))


class RunAllTasks(luigi.WrapperTask):
    """
    This class is a wrapper for the entire pipeline. It runs all the luigi tasks
    in the pipeline.
    """

    in_dir = Parameter(default=None)
    output_dir = Parameter(default=None)
    graph_dir = Parameter(default=None)
    first_month = Parameter(default=None)
    last_month = Parameter(default=None)
    meilisearch_url = Parameter(default=None)

    def requires(self):
        logging.info(f"Running all tasks with data until {self.last_month}...")
        try:
            # TODO Check if this is really needed
            if f"{self.last_month}_ocds_releases.json" not in os.listdir(
                self.output_dir
            ):
                shutil.rmtree(self.output_dir)
                shutil.rmtree(self.graph_dir)
                logging.warn(
                    f"Removed {self.graph_dir} and {self.output_dir} as all data needs to be reprocessed."
                )
        # If it's already gone for some reason, just don't error out
        except FileNotFoundError:
            pass

        return [
            # These are the tasks that no other task depends on
            BuildSearchIndex(
                meilisearch_url=self.meilisearch_url,
                in_dir=self.in_dir,
                output_dir=self.output_dir,
                first_month=self.first_month,
                last_month=self.last_month,
            ),
            LoadKuzuDB(
                graph_dir=self.graph_dir,
                in_dir=self.in_dir,
                output_dir=self.output_dir,
                first_month=self.first_month,
                last_month=self.last_month,
            ),
        ]


def run_pipeline() -> LuigiRunResult:
    """
    Run the TEDective ETL pipeline. This function serves as the entrypoint for
    running the entire pipeline via poetry scripts (defined in pyproject.toml).

    Args:
        first_month (str): The first month to process. Defaults to '2017-01'.
        last_month (str): The last month to process. Defaults to None.
        meilisearch_url (str): The URL of the Meilisearch server. Defaults to 'http://localhost:7700'.
        in_dir (str): The directory to store the TED XMLs. Defaults to '/tmp/ted_notices'.
        output_dir (str): The directory to store the output data. Defaults to '/tmp/output'.
        graph_dir (str): The name of the KuzuDB graph. Defaults to '/tmp/graph'.

    Returns:
        None
    """
    # HACK Provide all the args via the command line
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--first-month",
        help="The first month to process. Defaults to '2017-01'.",
        default="2017-01",
    )
    parser.add_argument(
        "--last-month",
        help="The last month to process. Defaults to the last month.",
        default=None,
    )
    parser.add_argument(
        "--meilisearch-url",
        help="The URL of the Meilisearch server. Defaults to 'http://localhost:7700'",
        default="http://localhost:7700",
    )
    parser.add_argument(
        "--in-dir",
        help="The directory to store the TED XMLs. Defaults to '/tmp/ted_notices'",
        default="/tmp/ted_notices",
    )
    parser.add_argument(
        "--output-dir",
        help="The directory to store the output data. Defaults to '/tmp/output'",
        default="/tmp/output",
    )
    parser.add_argument(
        "--graph-dir",
        help="The name of the KuzuDB graph. Defaults to '/tmp/graph'",
        default="/tmp/graph",
    )
    parser.add_argument(
        "--local-scheduler",
        help="Use the local scheduler.",
        action="store_true",
    )
    args = parser.parse_args()

    # If last_month is not provided, process TED XMLs until last complete
    # calendar month
    if args.last_month is None:
        logging.info("No end date provided. Processing TED XMLs until last month.")
        end_date = datetime.now() - relativedelta(months=1)
        args.last_month = end_date.strftime("%Y-%m")
    else:
        logging.info(f"Processing TED XMLs until {args.last_month}.")

    result = luigi.build(
        [
            RunAllTasks(
                first_month=str(args.first_month),
                last_month=str(args.last_month),
                meilisearch_url=str(args.meilisearch_url),
                in_dir=str(args.in_dir),
                output_dir=str(args.output_dir),
                graph_dir=str(args.graph_dir),
            ),
        ],
        detailed_summary=True,
        local_scheduler=args.local_scheduler,
    )
    if result.status == LuigiStatusCode.SUCCESS:
        logging.info("TEDective ETL pipeline completed successfully.")
        # Exit with status code 0
        exit(0)
    else:
        logging.error("TEDective ETL pipeline failed.")
        # Exit with status code 1
        exit(1)



# SPDX-FileCopyrightText: 2023 Free Software Foundation Europe <contact@fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

import os
import re
from datetime import date, datetime
from functools import cache
from pathlib import Path
from typing import Optional

import ftfy
import orjson
from _decimal import Decimal

from tedective_etl.schema import ocds


def get_project_root() -> Path:
    """
    Get the root directory of the project.

    Returns:
        Path: The root directory of the project.

    """
    if os.path.exists("/.dockerenv"):
        return Path("/app")
    return Path(__file__).parent.parent


def ocds_json_serializer(obj):
    """Custom JSON serializer for database insertion"""
    if isinstance(
        obj,
        (
            ocds.TagEnum,
            ocds.InitiationType,
            ocds.Status,
            ocds.ProcurementMethod,
            ocds.MainProcurementCategory,
            ocds.AwardStatus,
            ocds.ContractStatus,
            ocds.MilestoneStatus,
            ocds.Currency,
        ),
    ):
        return str(obj)
    # elif issubclass(type(obj), .SQLModel):
    # This allows quick and dirty insertion of JSON into models
    #    return obj.__dict__
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()


@cache
def clean(string: str) -> str:
    """
    Clean a string by removing extra whitespace and fixing unicode.

    Args:
        string (str): The string to clean.

    Returns:
        str: The cleaned string.

    """
    try:
        string = re.sub(r"\s+", " ", string)
        string = string.strip()
        string = ftfy.fix_text(string)
        return string
    except TypeError:
        return None


def find_closest_date(target_date, dates):
    # target_date = datetime.strptime(target_date, "%Y-%m-%d")
    closest_date = min(dates, key=lambda date: abs(datetime.strptime(date, "%Y-%m-%d") - target_date))
    return closest_date


def load_closest(date: str) ->dict:
    root = "/tmp" #get_project_root()
    path = os.path.join(root, "ted_currencies/rates.json")

    with open(path, "rb") as file:
        rates = orjson.loads(file.read())

    closest_date = find_closest_date(date, rates.keys()) 
    closest_rates= rates[closest_date]

    return closest_rates


def request_currency(amount: str, currency: str, date: str) -> Decimal:
    rates = load_closest(date)

    if currency in rates:
        rate = Decimal(rates[currency])
        amountEur = round(Decimal(amount) / rate)

        return amountEur
    else:
        return None


@cache
def convert_to_euro(amount: str, currency: str, date: str) -> Optional[Decimal]:
    return request_currency(amount, currency, date)


def prune_list(list_of_objects, seen_ids):
    """
    Prunes a list of objects by removing duplicates.

    Args:
        list_of_objects (list): The list of objects to prune.
        seen_ids (set): A set of ids that have already been seen.

    Returns:
        list: The pruned list of objects.
        set: The updated set of ids that have already been seen.

    """
    filtered_list = []
    ids = seen_ids
    for obj in list_of_objects:
        if type(obj) == ocds.Organization and obj.id not in ids:
            filtered_list.append(obj)
            ids.add(obj.id)
        elif type(obj) == ocds.Organization and obj.id in ids:
            continue
        else:
            filtered_list.append(obj)

    return list(set(filtered_list)), seen_ids



import logging
import os
import time

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq
from kuzu import Connection
from pyarrow import json  # noqa: F401


def create_release_schema(conn: Connection):
    """
    Creates a schema for the Release node in KuzuDB

    Args:
        conn (Connection): The database connection object.

    Returns:
        None
    """

    # TODO Add tag as STRING[] when this is fixed:
    # https://github.com/kuzudb/kuzu/issues/2703
    conn.execute(
        """
        CREATE NODE TABLE Release(
            id STRING, 
            ocid STRING, 
            date DATE,
            initiationType STRING,
            language STRING,
            tedURL STRING,
            PRIMARY KEY (id)
        )
        """
    )


def create_organization_schema(conn: Connection):
    """
    Creates a schema for the Organization node in KuzuDB

    Args:
        conn (Connection): The database connection object.

    Returns:
        None
    """

    # Add roles as STRING[], see above
    # Add additionalIdentifiers as STRUCT[], see above
    conn.execute(
        """
        CREATE NODE TABLE Organization(
            name STRING,
            id STRING,

            roles STRING,

            identifierNationalID STRING,

            addressStreetAddress STRING,
            addressLocality STRING,
            addressRegion STRING,
            addressPostalCode STRING,            
            addressCountryName STRING,

            contactPointName STRING,
            contactPointEmail STRING,
            contactPointTelephone STRING,
            contactPointFaxNumber STRING,
            contactPointUrl STRING,

            detailsUrl STRING,

            PRIMARY KEY (id)
        )
        """
    )


def create_tender_schema(conn: Connection):
    conn.execute(
        """
        CREATE NODE TABLE Tender(
            id STRING,
            title STRING, 
            description STRING, 
            status STRING,

            valueAmount DOUBLE,
            valueAmountEur DOUBLE,
            valueCurrency STRING,

            PRIMARY KEY (id)
        )
        """
    )


def create_award_schema(conn: Connection):
    conn.execute(
        """
        CREATE NODE TABLE Award(
            id STRING,
            title STRING, 
            status STRING,
            PRIMARY KEY (id)
        )
        """
    )


def create_contract_schema(conn: Connection):
    conn.execute(
        """
        CREATE NODE TABLE Contract(
            id STRING,
            title STRING,
            status STRING,

            valueAmount DOUBLE,
            valueAmountEur DOUBLE,
            valueCurrency STRING,

            dateSigned DATE,
            PRIMARY KEY (id)
        )
        """
    )


# TODO Create planning schema
def create_planning_schema(conn: Connection):
    pass


# TODO Create implementation schema
def create_implementation_schema(conn: Connection):
    pass


def create_edge_schema(conn: Connection):
    # From Release
    conn.execute("CREATE REL TABLE hasTender(FROM Release TO Tender)")
    conn.execute("CREATE REL TABLE hasBuyer(FROM Release TO Organization)")
    conn.execute("CREATE REL TABLE hasAwards(FROM Release TO Award)")
    conn.execute("CREATE REL TABLE hasContracts(FROM Release TO Contract)")
    # conn.execute("CREATE REL TABLE hasPlanning(FROM Release TO Planning)")

    # From Contract
    conn.execute("CREATE REL TABLE issuedAgainst(FROM Contract TO Award)")
    # conn.execute("CREATE REL TABLE hasImplementation(FROM Contract TO Implementation)")

    # To Organization
    conn.execute("CREATE REL TABLE hasSuppliers(FROM Award TO Organization)")
    conn.execute("CREATE REL TABLE hasTenderers(FROM Tender TO Organization)")
    conn.execute(
        "CREATE REL TABLE isSameAs(FROM Organization TO Organization, probability DOUBLE)"
    )


def create_kuzudb_schema(conn: Connection):
    """
    Creates the schema for KuzuDB

    Args:
        conn (Connection): The database connection object.

    Returns:
        None
    """
    # Create node schemas
    create_release_schema(conn)
    create_organization_schema(conn)
    create_tender_schema(conn)
    create_award_schema(conn)
    create_contract_schema(conn)
    # create_planning_schema(conn)
    # create_implementation_schema(conn)

    # Create edge schemas
    create_edge_schema(conn)


def create_import_files(
    path_parquet_data: str,
    path_organization_json: str,
    path_output: str,
):
    """
    Creates graph import files for KuzuDB

    Args:
        path_parquet_data (str): The path of to the parquet file containing the data.
        path_organization_json (str): The path of the JSON file containing the organization data.
        path_output (str): The path of the directory to store the import files.

    Returns:
        None
    """
    logging.info("Creating the import files for KuzuDB")

    # Read the parquet file
    logging.info("Reading the parquet file...")
    start = time.time()
    ocds_df = pd.read_parquet(path_parquet_data)
    end = time.time()
    logging.info(f"Time elapsed: {round(end - start, 2)} seconds")

    # Create the output directory if it doesn't exist
    if not os.path.exists(path_output):
        logging.info("Creating output directory")
        os.makedirs(path_output)

    # Create the output files
    logging.info("Creating output parquet files...")
    start = time.time()

    # 'Release' nodes
    logging.info("Creating nodes_releases.parquet...")
    release_cols = ["id", "ocid", "date", "initiationType", "language", "tedURL"]
    df = ocds_df[release_cols]
    df.loc[:, "date"] = pd.to_datetime(ocds_df["date"]).dt.date
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "nodes_releases.parquet"))

    # 'Organization' nodes
    logging.info("Creating nodes_organizations.parquet...")
    table = pa.json.read_json(path_organization_json)
    df = table.to_pandas()
    # Concatenate all the values in df['roles'] to a single string for now
    df["roles"] = df["roles"].apply(lambda x: " ".join(x))
    df["identifierNationalID"] = df["identifier"].apply(
        lambda x: x["id"] if pd.notnull(x) else ""
    )
    df["addressStreetAddress"] = df["address"].apply(
        lambda x: x["streetAddress"] if pd.notnull(x) else ""
    )
    df["addressLocality"] = df["address"].apply(
        lambda x: x["locality"] if pd.notnull(x) else ""
    )
    df["addressRegion"] = df["address"].apply(
        lambda x: x["region"] if pd.notnull(x) else ""
    )
    df["addressPostalCode"] = df["address"].apply(
        lambda x: x["postalCode"] if pd.notnull(x) else ""
    )
    df["addressCountryName"] = df["address"].apply(
        lambda x: x["countryName"] if pd.notnull(x) else ""
    )
    df["contactPointName"] = df["contactPoint"].apply(
        lambda x: x["name"] if pd.notnull(x) else ""
    )
    df["contactPointEmail"] = df["contactPoint"].apply(
        lambda x: x["email"] if pd.notnull(x) else ""
    )
    df["contactPointTelephone"] = df["contactPoint"].apply(
        lambda x: x["telephone"] if pd.notnull(x) else ""
    )
    df["contactPointFaxNumber"] = df["contactPoint"].apply(
        lambda x: x["faxNumber"] if pd.notnull(x) else ""
    )
    df["contactPointUrl"] = df["contactPoint"].apply(
        lambda x: x["url"] if pd.notnull(x) else ""
    )
    df["detailsUrl"] = df["details"].apply(lambda x: x["url"] if pd.notnull(x) else "")
    df = df.drop(
        columns=[
            "identifier",
            "additionalIdentifiers",
            "address",
            "contactPoint",
            "details",
        ]
    )
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(
        table,
        os.path.join(path_output, "nodes_organizations.parquet"),
    )

    # 'Tender' nodes
    logging.info("Creating nodes_tenders.parquet...")
    tender_cols = [
        "id",
        "title",
        "description",
        "status",
        "valueAmount",
        "valueAmountEur",
        "valueCurrency",
    ]
    df = pd.DataFrame(ocds_df["tender"].apply(lambda x: x if pd.notna(x) else None))
    df = pd.json_normalize(df["tender"].dropna(), max_level=0)
    df["valueAmount"] = df["value"].apply(
        lambda x: x["amount"] if pd.notna(x) else None
    )
    df["valueAmountEur"] = df["value"].apply(
        lambda x: x["amountEur"] if pd.notna(x) else None
    )
    df["valueCurrency"] = df["value"].apply(
        lambda x: x["currency"] if pd.notna(x) else None
    )
    table = pa.Table.from_pandas(df[tender_cols], preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "nodes_tenders.parquet"))

    # 'Award' nodes
    logging.info("Creating nodes_awards.parquet...")
    award_cols = ["id", "title", "status"]
    df = ocds_df.explode("awards", ignore_index=True)
    df = pd.json_normalize(df["awards"].dropna(), max_level=0)
    table = pa.Table.from_pandas(df[award_cols], preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "nodes_awards.parquet"))

    # 'Contract' nodes
    logging.info("Creating nodes_contracts.parquet...")
    contract_cols = [
        "id",
        "title",
        "status",
        "valueAmount",
        "valueAmountEur",
        "valueCurrency",
        "dateSigned",
    ]
    df = ocds_df.explode("contracts", ignore_index=True)
    df = pd.json_normalize(df["contracts"].dropna(), max_level=0)
    df["dateSigned"] = pd.to_datetime(df["dateSigned"]).dt.date
    df["valueAmount"] = df["value"].apply(
        lambda x: x["amount"] if pd.notna(x) else None
    )
    df["valueAmountEur"] = df["value"].apply(
        lambda x: x["amountEur"] if pd.notna(x) else None
    )
    df["valueCurrency"] = df["value"].apply(
        lambda x: x["currency"] if pd.notna(x) else None
    )
    table = pa.Table.from_pandas(df[contract_cols], preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "nodes_contracts.parquet"))

    # 'hasTender' edge (Release -> Tender)
    logging.info("Creating edges_hasTender.parquet...")
    release_ids = ocds_df["id"]
    tender_ids = ocds_df["tender"].apply(lambda x: x.get("id") if pd.notna(x) else None)
    df = pd.DataFrame({"from": release_ids, "to": tender_ids})
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "edges_hasTender.parquet"))

    # 'hasBuyer' edge (Release -> Organization)
    logging.info("Creating edges_hasBuyer.parquet...")
    buyer_ids = ocds_df["buyer"].apply(lambda x: x.get("id") if pd.notna(x) else None)
    df = pd.DataFrame({"from": release_ids, "to": buyer_ids})
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "edges_hasBuyer.parquet"))

    # 'hasAwards' edge (Release -> Award)
    logging.info("Creating edges_hasAwards.parquet...")
    df = ocds_df.explode("awards", ignore_index=True)
    df = df.dropna(subset=["awards"])
    df["from"] = df["id"]
    df["to"] = df["awards"].apply(lambda x: x["id"])
    df = df[["from", "to"]]
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "edges_hasAwards.parquet"))

    # 'hasContracts' edge (Award -> Contract)
    logging.info("Creating edges_hasContracts.parquet...")
    df = ocds_df.explode("contracts", ignore_index=True)
    df = df.dropna(subset=["contracts"])
    df["from"] = df["id"]
    df["to"] = df["contracts"].apply(lambda x: x["id"])
    df = df[["from", "to"]]
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "edges_hasContracts.parquet"))

    # 'hasSuppliers' edge (Award -> Organization)
    logging.info("Creating edges_hasSuppliers.parquet...")
    df = ocds_df.explode("awards", ignore_index=True)
    df = pd.json_normalize(df["awards"])
    df = df[["id", "suppliers"]]
    df = df.explode("suppliers", ignore_index=True)
    df = df.dropna(subset=["suppliers"])
    df["suppliers"] = df["suppliers"].apply(lambda x: x["id"])
    df = df.rename(columns={"id": "from", "suppliers": "to"}).reset_index(drop=True)
    table = pa.Table.from_pandas(df, preserve_index=False)
    pq.write_table(table, os.path.join(path_output, "edges_hasSuppliers.parquet"))

    end = time.time()
    logging.info(f"Time elapsed: {round(end - start, 2)} seconds")


def import_to_kuzu(path: str, conn: Connection):
    """
    Imports the data into KuzuDB

    Args:
        path (str): The path of the directory containing the import files.
        conn (Connection): The database connection object.

    Returns:
        None
    """
    # Start timer to track how long it takes to import data into KuzuDB
    logging.info("Importing data into KuzuDB...")
    start = time.time()

    # Import nodes
    logging.info("Importing Release nodes...")
    conn.execute(f"COPY Release FROM '{path}/nodes_releases.parquet'")
    logging.info("Importing organization nodes...")
    conn.execute(f"COPY Organization FROM '{path}/nodes_organizations.parquet'")
    logging.info("Importing Tender nodes...")
    conn.execute(f"COPY Tender FROM '{path}/nodes_tenders.parquet'")
    logging.info("Importing Award nodes...")
    conn.execute(f"COPY Award FROM '{path}/nodes_awards.parquet'")
    logging.info("Importing Contract nodes...")
    conn.execute(f"COPY Contract FROM '{path}/nodes_contracts.parquet'")

    # Import Edges
    logging.info("Importing edges...")
    conn.execute(f"COPY hasTender FROM '{path}/edges_hasTender.parquet'")
    conn.execute(f"COPY hasBuyer FROM '{path}/edges_hasBuyer.parquet'")
    conn.execute(f"COPY hasAwards FROM '{path}/edges_hasAwards.parquet'")
    conn.execute(f"COPY hasSuppliers FROM '{path}/edges_hasSuppliers.parquet'")
    conn.execute(f"COPY hasContracts FROM '{path}/edges_hasContracts.parquet'")
    conn.execute(f"COPY isSameAs FROM '{path}/edges_isSameAs.parquet'")

    # Stop timer
    end = time.time()
    logging.info(f"Time elapsed: {round(end - start, 2)} seconds")
    logging.info("Finished importing data into KuzuDB")



import logging
import os
import time

import orjson
from meilisearch import Client


def create_index(
    meilisearch_url: str = "http://localhost:7700",
    output_path: str = "/tmp/output",
    sleep_time: int = 7,
):
    # Create a Meilisearch client
    client = Client(meilisearch_url)

    # Delete indexes if they exist
    try:
        client.delete_index("organizations")
        client.delete_index("awards")
        client.delete_index("contracts")
    except Exception as e:
        logging.error(f"Error deleting indexes: {e}")

    logging.info("Existing indexes deleted")

    # Load file located at ../tedective-parser/output/organizations.jsonl into string
    # and then load the string into a list of dictionaries
    with open(os.path.join(output_path, "organizations.jsonl"), "r") as f:
        logging.info("Loading 'organizations' list into memory...")
        organizations = [orjson.loads(line) for line in f.readlines()]

    # Add documents to index
    logging.info("Adding 'organizations' to index...")
    index = client.index("organizations")
    index.update_filterable_attributes(
        [
            "address.locality",
            "address.region",
            "address.countryName",
            "roles",
            "details.classifications.id",
        ]
    )
    # Log size in MB of organizations list
    logging.debug(
        f"Size of 'organizations' list: {round(len(orjson.dumps(organizations)) / 1024 / 1024, 2)} MB"
    )
    # Chunk the list of organizations into smaller lists of 100000 items
    # and add each chunk to the index
    CHUNK_SIZE = 100000  # Make this a constant or a parameter
    for i in range(0, len(organizations), CHUNK_SIZE):
        try:
            index.add_documents(organizations[i : i + CHUNK_SIZE])
            logging.info(
                f"Chunk {(i // CHUNK_SIZE) + 1} from 'organizations' added to index tasks"
            )
            logging.info(
                f"Sleeping for {str(sleep_time)}s to give meilisearch some time to index..."
            )
            time.sleep(sleep_time)
        except Exception as e:
            logging.error(
                f"Failed to add 'organizations' chunk {i // CHUNK_SIZE} to index tasks: {e}"
            )

    # Load files located at ../tedective-parser/output/YYYYMM_ocds_releases.json into string
    for file in os.listdir(output_path):
        awards = []
        contracts = []
        if file.endswith(".json"):
            with open(os.path.join(output_path, file), "r") as f:
                # This is one big dictionary with a list of relevant
                # dictionaries under the key "releases"
                ocds = orjson.loads(f.read())

                # Get the list of releases
                releases = ocds["releases"]
                logging.debug(f"'{file}' loaded into memory")

                # Get the list of awards from each release
                for release in releases:
                    if release["awards"]:
                        for award in release["awards"]:
                            award = {k: v for k, v in award.items() if v}
                            awards.append(award)
                    if release["contracts"]:
                        for contract in release["contracts"]:
                            contract = {k: v for k, v in contract.items() if v}
                            contracts.append(contract)

                # Add awards to index
                index = client.index("awards")
                index.update_filterable_attributes(["status"])
                index.add_documents(awards, primary_key="id")
                logging.info(f"Awards from '{file}' added to 'awards' index")
                logging.info(
                    f"Sleeping for {str(sleep_time)}s to give meilisearch some time to index..."
                )
                time.sleep(sleep_time)

                # Add contracts to index
                index = client.index("contracts")
                index.update_filterable_attributes(["status", "value.currency"])
                index.update_sortable_attributes(["dateSigned", "value.amount"])
                index.add_documents(contracts, primary_key="id")
                logging.info(f"Contracts from '{file}' added to 'contracts' index")
                logging.info(
                    f"Sleeping for {str(sleep_time)}s to give meilisearch some time to index..."
                )
                time.sleep(sleep_time)
